/* Multirobot.c */
/* Pavel Surynek */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <assert.h>
#include <math.h>

#define CHECK_ERRORS		1

unsigned long long STAT_steps = 0;

//struct graph;

typedef struct vertex
{
	int m_val, m_low, m_2cc;
	struct vertex **m_nghs;
	struct vertex *m_f;
} VERTEX;


typedef struct stack_edge
{
	VERTEX *m_v1, *m_v2;
} STACK_EDGE;


typedef struct graph
{
	unsigned int m_V, m_E;
	VERTEX *m_vrtxs;
	VERTEX **m_edges;
} GRAPH;


typedef struct
{
	unsigned int m_R;
	VERTEX **m_vrtxs;
} MAPPING;


typedef struct move
{
	VERTEX *m_from;
	VERTEX *m_to;
	struct move *m_prev;
} MOVE;


typedef struct
{
	unsigned int m_V;
	int **m_d;
} DIST;


typedef struct state
{
	int m_q, m_d;
	int *m_vls;
	VERTEX **m_vrtxs;
	MOVE *m_mv;
} STATE;



typedef struct uvertex
{
	int m_id, m_val, m_low, m_art, m_cyc, m_rob, m_gob;
	struct uvertex *m_next;
	struct uedge *m_nghs;
	struct uvertex *m_father;
	struct uvertex *m_son;
	struct uvertex *m_lowv;
} UVERTEX;


typedef struct uedge
{
	int m_used;
	struct uvertex *m_v;
	struct uedge *m_next;
} UEDGE;


typedef struct ugraph
{
	unsigned int m_V, m_E;
	UVERTEX *m_root;
} UGRAPH;


typedef struct upool
{
	UVERTEX *m_vrtxs;
	UVERTEX *m_last_vrtx;
	UEDGE *m_edgs;
	UEDGE *m_last_edg;

	unsigned int m_V;
	unsigned int m_E;
	unsigned int m_maxV;
	unsigned int m_maxE;
} UPOOL;


typedef struct stack_uedge
{
	UVERTEX *m_v1, *m_v2;
} STACK_UEDGE;


typedef struct ustack
{
	unsigned int m_maxE;
	STACK_UEDGE *m_edgs;
	STACK_UEDGE *m_last;
} USTACK;


typedef struct cvertex
{
	UVERTEX *m_v;
	struct cvertex *m_next, *m_prev;
} CVERTEX;


typedef struct cycle
{
	unsigned int m_size;
	unsigned int m_back_size;
	CVERTEX *m_root;
	CVERTEX *m_back_root;	
	struct cycle *m_next;
	struct cycle *m_prev;
} CYCLE;


typedef struct cdeco
{
	unsigned int m_size;
	CYCLE *m_root;
	CYCLE *m_tail;
} CDECO;


typedef struct cpool
{
	CVERTEX *m_vrtxs;
	CVERTEX *m_last_vrtx;
	CYCLE *m_cycs;
	CYCLE *m_last_cyc;

	unsigned int m_V;
	unsigned int m_C;
	unsigned int m_maxV;
	unsigned int m_maxC;
} CPOOL;


typedef struct gvertex
{
	int m_id;
	UGRAPH m_2cc;
	CDECO m_cdeco;
	struct gvertex *m_next;
	struct gedge *m_nghs;
} GVERTEX;


typedef struct gedge
{
	struct gvertex *m_v;
	struct gedge *m_next;
} GEDGE;


typedef struct ggraph
{
	unsigned int m_V, m_E;
	GVERTEX *m_root;
} GGRAPH;


typedef struct gpool
{
	GVERTEX *m_vrtxs;
	GVERTEX *m_last_vrtx;
	GEDGE *m_edgs;
	GEDGE *m_last_edg;

	unsigned int m_V;
	unsigned int m_E;
	unsigned int m_maxV;
	unsigned int m_maxE;
} GPOOL;


typedef struct umove
{
	UVERTEX *m_from;
	UVERTEX *m_to;
	struct umove *m_prev;
} UMOVE;


VERTEX *BICON_vrtxs;
VERTEX *BICON_last_vrtx;
VERTEX **BICON_edgs;
VERTEX **BICON_last_edge;
GRAPH *BICON_comp;
GRAPH *BICON_last_comp;


STACK_EDGE *STACK_edges;
STACK_EDGE *STACK_last_edge;

int *STATE_vls;
VERTEX **STATE_vrtxs;

int *STATE_last_vl;
VERTEX **STATE_last_vrtx;

STATE *STATE_sts;
STATE **STATE_sts_heap;
STATE **STATE_last_st;

int *CLOSE_vls;
int *CLOSE_last_vl;

int *OPEN_vls;
int *OPEN_last_vl;

UGRAPH *last_2cc = NULL;

#define MULTIPLIER  1

#define STAR_QUEUE  (MULTIPLIER * 65536)
#define CLOSE_LIST  (MULTIPLIER * 65536)
#define OPEN_LIST   (MULTIPLIER * 65536)
#define SOLUTION    (MULTIPLIER * 65536)




typedef struct list_node
{
	struct list_node **m_nds;
} LIST_NODE;


LIST_NODE *CLOSE_root;
LIST_NODE *CLOSE_nds;
LIST_NODE *CLOSE_last_nd;

LIST_NODE **CLOSE_next_nds;
LIST_NODE **CLOSE_last_nnd;

LIST_NODE *OPEN_root;
LIST_NODE *OPEN_nds;
LIST_NODE *OPEN_last_nd;

LIST_NODE **OPEN_next_nds;
LIST_NODE **OPEN_last_nnd;

DIST dist_B;

MOVE solution[SOLUTION];
MOVE *SOL_last_mv;

MOVE sol_A[SOLUTION];
STATE TEMP_state;


void print_state(STATE *state, MAPPING *start, GRAPH *graph);
int get_quality_star(STATE *state, MAPPING *goal, GRAPH *graph);
void setup_distances(GRAPH *graph, DIST *dist);
void all_pairs_shortest(DIST *dist);
void up_heap(STATE **heap, STATE **last_state);
void down_heap(STATE **heap, STATE **last_state);

void print_closed(MAPPING *goal);
void close_state(STATE *state, MAPPING *goal, GRAPH *graph);
int is_closed(STATE *state, MAPPING *goal, GRAPH *graph);

void print_open(MAPPING *goal);
void open_state(STATE *state, MAPPING *goal, GRAPH *graph);
int is_open(STATE *state, MAPPING *goal, GRAPH *graph);

void close_state_tree(STATE *state, MAPPING *goal, GRAPH *graph);
int is_closed_tree(STATE *state, MAPPING *goal, GRAPH *graph);

void open_state_tree(STATE *state, MAPPING *goal, GRAPH *graph);
int is_open_tree(STATE *state, MAPPING *goal, GRAPH *graph);

void print_graph(GRAPH *graph);

void init_upool(UPOOL *upool, unsigned int maxV, unsigned int maxE);
void init_ustack(USTACK *ustack, unsigned int maxE);
void init_ugraph(UGRAPH *ugraph, UPOOL *upool);
void add_vertex_ugraph(int id, int r, int g, UGRAPH *ugraph, UPOOL *upool);
void add_edge_ugraph(int id1, int id2, UGRAPH *ugraph, UPOOL *upool);
void add_edge_ugraph_ptr(UVERTEX *v1, UVERTEX *v2, UGRAPH *ugraph, UPOOL *upool);
void mark_edge_ugraph(int id1, int id2, int used, UGRAPH *ugraph);
void print_ugraph(UGRAPH *ugraph);
void print_ustack(USTACK *ustack);
void read_ugraph(char *filename, UGRAPH *ugraph, UPOOL *upool);
void write_ugraph(char *filename, UGRAPH *ugraph);

void init_gpool(GPOOL *gpool, unsigned int maxV, unsigned int maxE);
void init_ggraph(GGRAPH *ggraph, GPOOL *gpool);
UGRAPH* add_vertex_ggraph(int id, GGRAPH *ggraph, GPOOL *gpool);
void add_edge_ggraph(int id1, int id2, GGRAPH *ggraph, GPOOL *gpool);
void add_edge_ggraph_ptr(GVERTEX *v1, GVERTEX *v2, GGRAPH *ggraph, GPOOL *gpool);
void print_ggraph(GGRAPH *ggraph);

void search_ucycle(UVERTEX *v, UGRAPH *ugraph, CDECO *cdeco, CPOOL *cpool);
void search_ucycle_root(UVERTEX *v, UGRAPH *ugraph, CDECO *cdeco, CPOOL *cpool);
void decompose_ucycle(UGRAPH *ugraph, CDECO *cdeco, CPOOL *cpool);
void decompose_ucycle1(UGRAPH *ugraph, CDECO *cdeco, CPOOL *cpool);

void init_cpool(CPOOL *cpool, unsigned int maxV, unsigned int maxC);
void init_cycle(CYCLE *cycle, CPOOL *cpool);
void init_cdeco(CDECO *cdeco);

void rotate_cycle_right(CYCLE *cycle, CDECO *cdeco);
void rotate_cycle_left(CYCLE *cycle, CDECO *cdeco);

int BICON_CNT = 0;
int COMP2_CNT = 0;


void init_cpool(CPOOL *cpool, unsigned int maxV, unsigned int maxC)
{
	cpool->m_vrtxs = (CVERTEX*)malloc(maxV * sizeof(CVERTEX));
	assert(cpool->m_vrtxs != NULL);
	cpool->m_last_vrtx = cpool->m_vrtxs;

	cpool->m_cycs = (CYCLE*)malloc(maxC * sizeof(CYCLE));
	assert(cpool->m_cycs != NULL);
	cpool->m_last_cyc = cpool->m_cycs;

	cpool->m_V = 0;
	cpool->m_C = 0;
	cpool->m_maxV = maxV;
	cpool->m_maxC = maxC;
}


void init_cycle(CYCLE *cycle, CPOOL *cpool)
{
	cycle->m_size = 0;
	cycle->m_back_size = 0;
	cycle->m_root = NULL;
	cycle->m_back_root = NULL;
	cycle->m_next = NULL;
}


void add_vertex_cycle(UVERTEX *v, CYCLE *cycle, CPOOL *cpool)
{
	CVERTEX **u = &cycle->m_root;
	CVERTEX *prev = NULL;

	while (*u != NULL)
	{
		prev = *u;
		u = &(*u)->m_next;
	}

	++cpool->m_V;
	assert(cpool->m_V <= cpool->m_maxV);

	*u = cpool->m_last_vrtx++;
	(*u)->m_v = v;
	(*u)->m_next = NULL;
	(*u)->m_prev = prev;

	++cycle->m_size;
}


void add_back_vertex_cycle(UVERTEX *v, CYCLE *cycle, CPOOL *cpool)
{
	CVERTEX **u = &cycle->m_back_root;
	CVERTEX *prev = NULL;

	while (*u != NULL)
	{
		prev = *u;
		u = &(*u)->m_next;
	}

	++cpool->m_V;
	assert(cpool->m_V <= cpool->m_maxV);

	*u = cpool->m_last_vrtx++;
	(*u)->m_v = v;
	(*u)->m_next = NULL;
	(*u)->m_prev = prev;

	++cycle->m_back_size;
}


void add_cycle_cdeco(CYCLE *cycle, CDECO *cdeco, CPOOL *cpool)
{
	CYCLE *last = NULL;
	CYCLE **c = &cdeco->m_root;

	while (*c != NULL)
	{
		last = *c;
		c = &(*c)->m_next;
	}

	++cpool->m_C;
	assert(cpool->m_C <= cpool->m_maxC);

	(*c) = cpool->m_last_cyc++;
	(*c)->m_size = cycle->m_size;
	(*c)->m_back_size = cycle->m_back_size;
	(*c)->m_root = cycle->m_root;
	(*c)->m_back_root = cycle->m_back_root;
	(*c)->m_next = NULL;
	(*c)->m_prev = last;

	++cdeco->m_size;
	cdeco->m_tail = *c;
}


void print_cycle(CYCLE *cycle)
{
	printf("Cycle:%d\n", cycle->m_size);

	CVERTEX *u = cycle->m_root;

	while (u != NULL)
	{
		printf("%d ", u->m_v->m_id); 
		u = u->m_next;
	}
	printf("\n");

	CVERTEX *w = cycle->m_back_root;

	while (w != NULL)
	{
		printf("%d ", w->m_v->m_id); 
		w = w->m_next;
	}
	printf("\n");
}


void print_cdeco(CDECO *cdeco)
{
	printf("Cycle decomposition:%d\n", cdeco->m_size);

	CYCLE *c = cdeco->m_root;

	while (c != NULL)
	{
		print_cycle(c);
		c = c->m_next;
	}
}


void init_cdeco(CDECO *cdeco)
{
	cdeco->m_size = 0;
	cdeco->m_root = NULL;
	cdeco->m_tail = NULL;
}


void init_upool(UPOOL *upool, unsigned int maxV, unsigned int maxE)
{
	upool->m_vrtxs = (UVERTEX*)malloc(maxV * sizeof(UVERTEX));
	assert(upool->m_vrtxs != NULL);
	upool->m_last_vrtx = upool->m_vrtxs;

	upool->m_edgs = (UEDGE*)malloc(maxE * sizeof(UEDGE));
	assert(upool->m_edgs != NULL);
	upool->m_last_edg = upool->m_edgs;

	upool->m_V = 0;
	upool->m_E = 0;
	upool->m_maxV = maxV;
	upool->m_maxE = maxE;
}


void init_ustack(USTACK *ustack, unsigned int maxE)
{
	ustack->m_edgs = (STACK_UEDGE*)malloc(maxE * sizeof(STACK_EDGE));
	assert(ustack->m_edgs != NULL);
	ustack->m_last = ustack->m_edgs - 1;
	ustack->m_maxE = maxE;
}


void push_ustack(UVERTEX *v1, UVERTEX *v2, USTACK *ustack)
{
	++ustack->m_last;
	assert(ustack->m_last - ustack->m_edgs <= ustack->m_maxE);
	ustack->m_last->m_v1 = v1;
	ustack->m_last->m_v2 = v2;
}


void pop_ustack(USTACK *ustack)
{
	assert(ustack->m_last >= ustack->m_edgs);
	--ustack->m_last;
}


void print_ustack(USTACK *ustack)
{
	STACK_UEDGE *e = ustack->m_edgs;

	printf("Undirected stack\n");

	while (e <= ustack->m_last)
	{
		printf("%d %d\n", e->m_v1->m_id, e->m_v2->m_id);
		++e;
	}
}


void init_ugraph(UGRAPH *ugraph, UPOOL *upool)
{
	ugraph->m_V = 0;
	ugraph->m_E = 0;
	ugraph->m_root = NULL;
}


void add_vertex_ugraph(int id, int r, int g, UGRAPH *ugraph, UPOOL *upool)
{
	++ugraph->m_V;
	++upool->m_V;
	assert(upool->m_V <= upool->m_maxV);

	UVERTEX **v = &ugraph->m_root;

	while (*v != NULL)
	{
		if ((*v)->m_id == id)
		{
			return;
		}
		v = &(*v)->m_next;
	}

	*v = upool->m_last_vrtx++;

	(*v)->m_id = id;
	(*v)->m_rob = r;
	(*v)->m_gob = g;
	(*v)->m_next = NULL;
	(*v)->m_nghs = NULL;
}


void add_edge_ugraph(int id1, int id2, UGRAPH *ugraph, UPOOL *upool)
{
	UVERTEX *v = ugraph->m_root;
	UVERTEX *u1, *u2;

	u1 = u2 = NULL;

	while (v != NULL)
	{
		if (v->m_id == id1)
		{
			u1 = v;

			if (u2 != NULL)
			{
				break;
			}
		}
		if (v->m_id == id2)
		{
			u2 = v;

			if (u1 != NULL)
			{
				break;
			}
		}
		v = v->m_next;
	}
	add_edge_ugraph_ptr(u1, u2, ugraph, upool);
}


void mark_edge_ugraph(int id1, int id2, int used, UGRAPH *ugraph)
{
	UVERTEX *v = ugraph->m_root;
	UVERTEX *u1, *u2;

	u1 = u2 = NULL;

	while (v != NULL)
	{
		if (v->m_id == id1)
		{
			u1 = v;

			if (u2 != NULL)
			{
				break;
			}
		}
		if (v->m_id == id2)
		{
			u2 = v;

			if (u1 != NULL)
			{
				break;
			}
		}
		v = v->m_next;
	}

	UEDGE **e1 = &u1->m_nghs;

	while (*e1 != NULL)
	{
		if ((*e1)->m_v == u2)
		{
			break;
		}
		e1 = &(*e1)->m_next;
	}

	UEDGE **e2 = &u2->m_nghs;

	while (*e2 != NULL)
	{
		if ((*e2)->m_v == u1)
		{
			break;
		}
		e2 = &(*e2)->m_next;
	}
	(*e1)->m_used = used;
	(*e2)->m_used = used;
}


void add_edge_ugraph_ptr(UVERTEX *v1, UVERTEX *v2, UGRAPH *ugraph, UPOOL *upool)
{
	UEDGE **e1 = &v1->m_nghs;

	while (*e1 != NULL)
	{
		if ((*e1)->m_v == v2)
		{
			break;
		}
		e1 = &(*e1)->m_next;
	}
	if (*e1 == NULL)
	{
		++ugraph->m_E;
		++upool->m_E;
		assert(upool->m_E <= upool->m_maxE);
		*e1 = upool->m_last_edg++;

		(*e1)->m_v = v2;
		(*e1)->m_next = NULL;
		(*e1)->m_used = 0;
	}

	UEDGE **e2 = &v2->m_nghs;

	while (*e2 != NULL)
	{
		if ((*e2)->m_v == v1)
		{
			break;
		}
		e2 = &(*e2)->m_next;
	}
	if (*e2 == NULL)
	{
		++ugraph->m_E;
		++upool->m_E;
		assert(upool->m_E <= upool->m_maxE);
		*e2 = upool->m_last_edg++;

		(*e2)->m_v = v1;
		(*e2)->m_next = NULL;
		(*e2)->m_used = 0;
	}
}


void print_ugraph(UGRAPH *ugraph)
{
	UEDGE *n;
	UVERTEX *v = ugraph->m_root;

	printf("Undirected graph\n");
	
	while (v != NULL)
	{
		printf("%d (val:%d low:%d art:%d r:%d g:%d):", v->m_id, v->m_val, v->m_low, v->m_art, v->m_rob, v->m_gob);
		n = v->m_nghs;

		while (n != NULL)
		{
			printf("%d ", n->m_v->m_id);
			n = n->m_next;
		}
		printf("\n");

		v = v->m_next;
	}
}


void read_ugraph(char *filename, UGRAPH *ugraph, UPOOL *upool)
{
	FILE *fr;
	unsigned int i, V, E, u, v;

	fr = fopen(filename, "r");

#ifdef CHECK_ERRORS
	if (fr == NULL)
	{
		printf("Error 21: Cannot open file %s.\n", filename);
		exit(-13);
	}
#endif

	fscanf(fr, "p edge %u %u\n", &V, &E);
	init_upool(upool, V, 2 * E);

	for (i = 0; i < V; ++i)
	{
		add_vertex_ugraph(i, -1, -1, ugraph, upool);
	}

	for (i = 0; i < E; ++i)
	{
		fscanf(fr, "e %u %u\n", &u, &v);
		add_edge_ugraph(u, v, ugraph, upool);
	}

	fclose(fr);
}


void write_ugraph(char *filename, UGRAPH *ugraph)
{
	FILE *fw;

	fw = fopen(filename, "w");

#ifdef CHECK_ERRORS
	if (fw == NULL)
	{
		printf("Error 15: Cannot open file %s for writing.\n", filename);
		exit(-15);
	}
#endif

	fprintf(fw, "p edge %u %u\n", ugraph->m_V, ugraph->m_E / 2);

	UEDGE *n;
	UVERTEX *v = ugraph->m_root;
	
	while (v != NULL)
	{
		v->m_id;
		n = v->m_nghs;

		while (n != NULL)
		{
			if (v->m_id < n->m_v->m_id)
			{
				fprintf(fw, "e %u %u\n", v->m_id, n->m_v->m_id);
			}

			n = n->m_next;
		}
		v = v->m_next;
	}

	fclose(fw);
}


void init_gpool(GPOOL *gpool, unsigned int maxV, unsigned int maxE)
{
	gpool->m_vrtxs = (GVERTEX*)malloc(maxV * sizeof(GVERTEX));
	assert(gpool->m_vrtxs != NULL);
	gpool->m_last_vrtx = gpool->m_vrtxs;

	gpool->m_edgs = (GEDGE*)malloc(maxE * sizeof(GEDGE));
	assert(gpool->m_edgs != NULL);
	gpool->m_last_edg = gpool->m_edgs;

	gpool->m_V = 0;
	gpool->m_E = 0;
	gpool->m_maxV = maxV;
	gpool->m_maxE = maxE;
}


void init_ggraph(GGRAPH *ggraph, GPOOL *gpool)
{
	ggraph->m_V = 0;
	ggraph->m_E = 0;
	ggraph->m_root = NULL;
}


UGRAPH* add_vertex_ggraph(int id, GGRAPH *ggraph, GPOOL *gpool)
{
	++ggraph->m_V;
	++gpool->m_V;
	assert(gpool->m_V <= gpool->m_maxV);

	GVERTEX **v = &ggraph->m_root;

	while (*v != NULL)
	{
		if ((*v)->m_id == id)
		{
			return &(*v)->m_2cc;
		}
		v = &(*v)->m_next;
	}

	*v = gpool->m_last_vrtx++;

	(*v)->m_id = id;
	(*v)->m_next = NULL;
	(*v)->m_nghs = NULL;

	return &(*v)->m_2cc;
}


void add_edge_ggraph(int id1, int id2, GGRAPH *ggraph, GPOOL *gpool)
{
	GVERTEX *v = ggraph->m_root;
	GVERTEX *u1, *u2;

	u1 = u2 = NULL;

	while (v != NULL)
	{
		if (v->m_id == id1)
		{
			u1 = v;

			if (u2 != NULL)
			{
				break;
			}
		}
		if (v->m_id == id2)
		{
			u2 = v;

			if (u1 != NULL)
			{
				break;
			}
		}
		v = v->m_next;
	}
	add_edge_ggraph_ptr(u1, u2, ggraph, gpool);
}


void add_edge_ggraph_ptr(GVERTEX *v1, GVERTEX *v2, GGRAPH *ggraph, GPOOL *gpool)
{
	GEDGE **e1 = &v1->m_nghs;

	while (*e1 != NULL)
	{
		if ((*e1)->m_v == v2)
		{
			break;
		}
		e1 = &(*e1)->m_next;
	}
	if (*e1 == NULL)
	{
		++ggraph->m_E;
		++gpool->m_E;
		assert(gpool->m_E <= gpool->m_maxE);
		*e1 = gpool->m_last_edg++;

		(*e1)->m_v = v2;
		(*e1)->m_next = NULL;
	}

	GEDGE **e2 = &v2->m_nghs;

	while (*e2 != NULL)
	{
		if ((*e2)->m_v == v1)
		{
			break;
		}
		e2 = &(*e2)->m_next;
	}
	if (*e2 == NULL)
	{
		++ggraph->m_E;
		++gpool->m_E;
		assert(gpool->m_E <= gpool->m_maxE);
		*e2 = gpool->m_last_edg++;

		(*e2)->m_v = v1;
		(*e2)->m_next = NULL;
	}
}


void print_ggraph(GGRAPH *ggraph)
{
	GEDGE *n;
	GVERTEX *v = ggraph->m_root;

	printf("Biconnected-component graph\n");
	
	while (v != NULL)
	{
		printf("%d (2cc:%p):", v->m_id, &v->m_2cc);
		n = v->m_nghs;

		while (n != NULL)
		{
			printf("%d ", n->m_v->m_id);
			n = n->m_next;
		}
		printf("\n");
		print_ugraph(&v->m_2cc);
		print_cdeco(&v->m_cdeco);

		v = v->m_next;
	}
}


void solve_ggraph(GGRAPH *ggraph)
{
	GEDGE *n;
	GVERTEX *v = ggraph->m_root;
	
	while (v != NULL)
	{
		n = v->m_nghs;

		while (n != NULL)
		{
			n = n->m_next;
		}

		CYCLE *c = v->m_cdeco.m_root;
		
		while (c != NULL)
		{
			rotate_cycle_left(c, &v->m_cdeco);
			c = c->m_next;
		}
		v = v->m_next;
	}
}


void search_ubiconnected(UVERTEX *v, UGRAPH *ugraph, USTACK *ustack, GGRAPH *tree, UPOOL *upool, GPOOL *gpool)
{
	UGRAPH *ug;

	v->m_val = BICON_CNT++;
	v->m_low = v->m_val;

	UEDGE *n;
	
	for (n = v->m_nghs; n != NULL; n = n->m_next)
	{
		if (n->m_used == 0)
		{
			mark_edge_ugraph(v->m_id, n->m_v->m_id, 1, ugraph);

			push_ustack(v, n->m_v, ustack);
			printf("push:%d,%d\n", v->m_id, n->m_v->m_id);

			if (n->m_v->m_low == -1)
			{
				n->m_v->m_father = v;
				v->m_son = n->m_v;
				search_ubiconnected(n->m_v, ugraph, ustack, tree, upool, gpool);
				
				if (n->m_v->m_low >= v->m_val && v->m_val != 0)
				{
					printf("--> arti:%d <--\n", v->m_id);
					v->m_art = 1;
					printf("beta 1\n");
					last_2cc = ug = add_vertex_ggraph(v->m_id, tree, gpool);
					printf("beta 2\n");
					init_ugraph(ug, upool);
					printf("beta 3\n");
					print_ustack(ustack);
					printf("beta 4\n");

					while (ustack->m_last > ustack->m_edgs)
					{
						printf("beta 9\n");
						if (ustack->m_last->m_v1->m_art != 0)
						{
							if (ustack->m_last->m_v2->m_art != 0)
							{
								printf("** Bridge:%d,%d\n", ustack->m_last->m_v1->m_id, ustack->m_last->m_v2->m_id);
								add_edge_ggraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v2->m_id, tree, gpool);
							}
							else
							{
								if (ustack->m_last->m_v1 == v)
								{
									printf("beta 5\n");
									add_vertex_ugraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v1->m_rob, ustack->m_last->m_v1->m_gob, ug, upool);
									printf("beta 6\n");
									add_vertex_ugraph(ustack->m_last->m_v2->m_id, ustack->m_last->m_v2->m_rob, ustack->m_last->m_v2->m_gob, ug, upool);
									printf("beta 7\n");
									add_edge_ugraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v2->m_id, ug, upool);
								}
							}
						}
						else
						{
							if (ustack->m_last->m_v2->m_art != 0)
							{
								if (ustack->m_last->m_v2 == v)
								{
									printf("beta 5\n");
									add_vertex_ugraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v1->m_rob, ustack->m_last->m_v1->m_gob, ug, upool);
									printf("beta 6\n");
									add_vertex_ugraph(ustack->m_last->m_v2->m_id, ustack->m_last->m_v2->m_rob, ustack->m_last->m_v2->m_gob, ug, upool);
									printf("beta 7\n");
									add_edge_ugraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v2->m_id, ug, upool);
								}
							}
							else
							{
								printf("beta 5\n");
								add_vertex_ugraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v1->m_rob, ustack->m_last->m_v1->m_gob, ug, upool);
								printf("beta 6\n");
								add_vertex_ugraph(ustack->m_last->m_v2->m_id, ustack->m_last->m_v2->m_rob, ustack->m_last->m_v2->m_gob, ug, upool);
								printf("beta 7\n");
								add_edge_ugraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v2->m_id, ug, upool);
							}
						}
						if (v == ustack->m_last->m_v1 && n->m_v == ustack->m_last->m_v2)
						{
							break;
						}
						
						printf("pop:%d,%d\n", ustack->m_last->m_v1->m_id, ustack->m_last->m_v2->m_id);
						pop_ustack(ustack);
						printf("beta 8\n");
						
					}
					if (ustack->m_last > ustack->m_edgs)
					{
						printf("pop:%d,%d\n", ustack->m_last->m_v1->m_id, ustack->m_last->m_v2->m_id);
						pop_ustack(ustack);
					}
					++COMP2_CNT;
					printf("--> Delta 1\n");
					print_ustack(ustack);
					printf("--> Delta 2\n");

				}
				v->m_low = n->m_v->m_low < v->m_low ? n->m_v->m_low : v->m_low;
			}
			else
			{
				if (n->m_v->m_father != v->m_father)
				{
					v->m_low = n->m_v->m_val < v->m_low ? n->m_v->m_val : v->m_low;
				}
			}
		}
	}
}


void search_ubiconnected_root(UVERTEX *v, UGRAPH *ugraph, USTACK *ustack, GGRAPH *tree, UPOOL *upool, GPOOL *gpool)
{
	UGRAPH *ug = NULL;

	v->m_val = BICON_CNT++;
	v->m_low = v->m_val;

	UEDGE *n;
	
	for (n = v->m_nghs; n != NULL; n = n->m_next)
	{
		if (n->m_used == 0)
		{
			mark_edge_ugraph(v->m_id, n->m_v->m_id, 1, ugraph);
			push_ustack(v, n->m_v, ustack);

			if (n->m_v->m_low == -1)
			{
				n->m_v->m_father = v;
				v->m_son = n->m_v;
				search_ubiconnected(n->m_v, ugraph, ustack, tree, upool, gpool);
				
				if (n->m_v->m_low >= v->m_val && v->m_val != 0)
				{
					v->m_art = 1;
					last_2cc = ug = add_vertex_ggraph(v->m_id, tree, gpool);
					init_ugraph(ug, upool);
					printf("--> Delta 3\n");
					print_ustack(ustack);
					printf("--> Delta 4\n");

					while (ustack->m_last > ustack->m_edgs)
					{
						if (ustack->m_last->m_v1->m_art != 0)
						{
							if (ustack->m_last->m_v2->m_art != 0)
							{
								add_edge_ggraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v2->m_id, tree, gpool);
							}
							else
							{
								if (ustack->m_last->m_v1 == v)
								{
									add_vertex_ugraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v1->m_rob, ustack->m_last->m_v1->m_gob, ug, upool);
									add_vertex_ugraph(ustack->m_last->m_v2->m_id, ustack->m_last->m_v2->m_rob, ustack->m_last->m_v2->m_gob, ug, upool);
									add_edge_ugraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v2->m_id, ug, upool);
								}
							}
						}
						else
						{
							if (ustack->m_last->m_v2->m_art != 0)
							{
								if (ustack->m_last->m_v2 == v)
								{
									add_vertex_ugraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v1->m_rob, ustack->m_last->m_v1->m_gob,ug, upool);
									add_vertex_ugraph(ustack->m_last->m_v2->m_id, ustack->m_last->m_v2->m_rob, ustack->m_last->m_v2->m_gob, ug, upool);
									add_edge_ugraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v2->m_id, ug, upool);
								}
							}
							else
							{
								add_vertex_ugraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v1->m_rob, ustack->m_last->m_v1->m_gob, ug, upool);
								add_vertex_ugraph(ustack->m_last->m_v2->m_id, ustack->m_last->m_v1->m_rob, ustack->m_last->m_v1->m_gob, ug, upool);
								add_edge_ugraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v2->m_id, ug, upool);
							}
						}
						if (v == ustack->m_last->m_v1 && n->m_v == ustack->m_last->m_v2)
						{
							break;
						}
						pop_ustack(ustack);
						
					}
					if (ustack->m_last > ustack->m_edgs)
					{
						pop_ustack(ustack);
					}
					++COMP2_CNT;
					printf("--> Delta 5\n");
					print_ustack(ustack);
					printf("--> Delta 6\n");

				}
				v->m_low = n->m_v->m_low < v->m_low ? n->m_v->m_low : v->m_low;
			}
			else
			{
				if (n->m_v->m_father != v->m_father)
				{
					v->m_low = n->m_v->m_val < v->m_low ? n->m_v->m_val : v->m_low;
				}
			}
		}
	}

	if (COMP2_CNT > 0)
	{
		--COMP2_CNT;
	}

	printf("ug:%p\n", ug);
//	print_ugraph(ug);
	if (last_2cc == NULL)
	{
		last_2cc = add_vertex_ggraph(v->m_id, tree, gpool);
		init_ugraph(last_2cc, upool);
	}

	while (ustack->m_last >= ustack->m_edgs)
	{
		if (ustack->m_last->m_v1->m_art != 0 && ustack->m_last->m_v2->m_art != 0)
		{
			printf("** Bridge:%d,%d\n", ustack->m_last->m_v1->m_id, ustack->m_last->m_v2->m_id);
			add_edge_ggraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v2->m_id, tree, gpool);
		}
		else
		{
			add_vertex_ugraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v1->m_rob, ustack->m_last->m_v1->m_gob, last_2cc, upool);
			add_vertex_ugraph(ustack->m_last->m_v2->m_id, ustack->m_last->m_v2->m_rob, ustack->m_last->m_v2->m_gob, last_2cc, upool);
			add_edge_ugraph(ustack->m_last->m_v1->m_id, ustack->m_last->m_v2->m_id, last_2cc, upool);
		}
		
		printf("pop:%d,%d\n", ustack->m_last->m_v1->m_id, ustack->m_last->m_v2->m_id);
		pop_ustack(ustack);
		
	}
	printf("--> Delta 7\n");
	print_ustack(ustack);
	printf("--> Delta 8\n");
}

 
void decompose_ubiconnected(UGRAPH *ugraph, USTACK *ustack, GGRAPH *tree, UPOOL *upool, GPOOL *gpool, CPOOL *cpool)
{
	UVERTEX *v;

	printf("***---> Gamma 1\n");
	init_ustack(ustack, ugraph->m_E);

	init_gpool(gpool, ugraph->m_V, ugraph->m_E);
	init_upool(upool, ugraph->m_V + ugraph->m_E, ugraph->m_V + ugraph->m_E);
	init_cpool(cpool, ugraph->m_V + ugraph->m_E, ugraph->m_V + ugraph->m_E);
	init_ggraph(tree, gpool);

	for (v = ugraph->m_root; v != NULL; v = v->m_next)
	{
		v->m_low = -1;
		v->m_lowv = NULL;
		v->m_art = 0;
		v->m_son = NULL;
	}

	for (v = ugraph->m_root; v != NULL; v = v->m_next)
	{
		if (v->m_low == -1)
		{
			v->m_father = NULL;
			search_ubiconnected_root(v, ugraph, ustack, tree, upool, gpool);
		}
	}

	GEDGE *n;
	GVERTEX *u = tree->m_root;

	printf("***---> Gamma 2\n");

	while (u != NULL)
	{
		if (u->m_2cc.m_V > 0)
		{
			init_cdeco(&u->m_cdeco);
			printf("***---> Gamma 3\n");
			decompose_ucycle1(&u->m_2cc, &u->m_cdeco, cpool);
			printf("***---> Gamma 4\n");
		}
		u = u->m_next;
	}
	printf("***---> Gamma 5\n");

//	destroy_stack();
}


int cc = 0;

void search_ucycle(UVERTEX *v, UGRAPH *ugraph, CDECO *cdeco, CPOOL *cpool)
{
	UGRAPH *ug;

	v->m_val = BICON_CNT++;
	v->m_low = v->m_val;
	v->m_lowv = v;

	UEDGE *n;
	
	for (n = v->m_nghs; n != NULL; n = n->m_next)
	{
		if (n->m_used == 0)
		{
			mark_edge_ugraph(v->m_id, n->m_v->m_id, 1, ugraph);

			if (n->m_v->m_low == -1)
			{
				n->m_v->m_father = v;
				v->m_son = n->m_v;
				search_ucycle(n->m_v, ugraph, cdeco, cpool);
				
				if (v->m_low > n->m_v->m_low)
				{
					v->m_low = n->m_v->m_low;
					v->m_lowv = n->m_v->m_lowv;
				}
			}
			else
			{
				if (n->m_v->m_father != v->m_father)
				{
					if (v->m_low > n->m_v->m_low)
					{
						v->m_low = n->m_v->m_low;
						v->m_lowv = n->m_v->m_lowv;
					}
				}
			}
		}
	}
}


void search_ucycle_root(UVERTEX *v, UGRAPH *ugraph, CDECO *cdeco, CPOOL *cpool)
{
	UGRAPH *ug = NULL;

	v->m_val = BICON_CNT++;
	v->m_low = v->m_val;
	v->m_lowv = v;

	UEDGE *n;

	for (n = v->m_nghs; n != NULL; n = n->m_next)
	{
		if (n->m_used == 0)
		{
			mark_edge_ugraph(v->m_id, n->m_v->m_id, 1, ugraph);

			if (n->m_v->m_low == -1)
			{
				n->m_v->m_father = v;
				v->m_son = n->m_v;
				search_ucycle(n->m_v, ugraph, cdeco, cpool);

				if (v->m_low > n->m_v->m_low)
				{
					v->m_low = n->m_v->m_low;
					v->m_lowv = n->m_v->m_lowv;
				}
			}
			else
			{
				if (n->m_v->m_father != v->m_father)
				{
					if (v->m_low > n->m_v->m_low)
					{
						v->m_low = n->m_v->m_low;
						v->m_lowv = n->m_v->m_lowv;
					}
				}
			}
		}
	}
}


UVERTEX* shortest_path(UVERTEX *v, UGRAPH *ugraph)
{
	UVERTEX *u, *l, *x;

	x = v;

	for (u = ugraph->m_root; u != NULL; u = u->m_next)
	{
		u->m_low = ugraph->m_V + 1;
		u->m_father = NULL;
		u->m_art = 0;
	}

	v->m_low = 0;

	while (v != NULL)
	{
		v->m_art = 1;
		UEDGE *n;

		for (n = v->m_nghs; n != NULL; n = n->m_next)
		{
			if (n->m_v->m_art == 0)
			{
				if (v->m_cyc == 0 || n->m_v->m_cyc == 0)
				{
					if (n->m_v->m_low > v->m_low + 1)
					{
						printf("x:%d\n", n->m_v->m_id);
						n->m_v->m_father = v;
						n->m_v->m_low = v->m_low + 1;
					
						if (n->m_v->m_cyc != 0 && n->m_v != x)
						{
							return n->m_v;
						}
					}
				}
			}
		}

		l = NULL;

		for (u = ugraph->m_root; u != NULL; u = u->m_next)
		{
			if (u->m_art == 0)
			{
				if (l == NULL)
				{
					l = u;
				}
				else
				{
					if (u->m_low < l->m_low)
					{
						l = u;
					}
				}
			}
		}
		v = l;
	}
	printf("NIL\n");
	print_ugraph(ugraph);
	return NULL;
}


UVERTEX* shortest_back_path(UVERTEX *v1, UVERTEX *v2, UGRAPH *ugraph)
{
	UVERTEX *u, *l, *x;

	x = v1;

	for (u = ugraph->m_root; u != NULL; u = u->m_next)
	{
		u->m_low = ugraph->m_V + 1;
		u->m_son = NULL;
		u->m_art = 0;
	}

	v1->m_low = 0;

	while (v1 != NULL)
	{
		v1->m_art = 1;
		UEDGE *n;

		for (n = v1->m_nghs; n != NULL; n = n->m_next)
		{
			if (n->m_v->m_art == 0)
			{
				if (n->m_v->m_cyc != 0)
				{
					if (n->m_v->m_low > v1->m_low + 1)
					{
						n->m_v->m_son = v1;
						n->m_v->m_low = v1->m_low + 1;
					
						if (n->m_v == v2)
						{
							printf("bp:%d,%d\n", v1->m_id, n->m_v->m_id);
							return n->m_v;
						}
					}
				}
			}
		}

		l = NULL;

		for (u = ugraph->m_root; u != NULL; u = u->m_next)
		{
			if (u->m_art == 0)
			{
				if (l == NULL)
				{
					l = u;
				}
				else
				{
					if (u->m_low < l->m_low)
					{
						l = u;
					}
				}
			}
		}
		v1 = l;
	}
	printf("NIL NIL\n");
	print_ugraph(ugraph);
	return NULL;
}


void decompose_ucycle1(UGRAPH *ugraph, CDECO *cdeco, CPOOL *cpool)
{
	UVERTEX *v, *u, *w, *w1, *w2;

	cc = 0;

	for (v = ugraph->m_root; v != NULL; v = v->m_next)
	{
		v->m_low = -1;
		v->m_art = 0;
		v->m_cyc = 0;
		v->m_son = NULL;
	}

	for (v = ugraph->m_root; v != NULL; v = v->m_next)
	{
		if (v->m_cyc == 0)
		{
			CYCLE cycle;
			init_cycle(&cycle, cpool);

			++cc;

			UEDGE *n, *m, *o;
	
			for (n = v->m_nghs; n != NULL; n = n->m_next)
			{
				if (n->m_v->m_cyc == 0)
				{
					break;
				}
			}

			if (n != NULL)
			{
				n->m_v->m_father = v;

				for (m = v->m_nghs; m != NULL; m = m->m_next)
				{
					if (m->m_v->m_cyc != 0)
					{
						break;
					}
				}

				u = n->m_v;

				v->m_cyc = cc;
				u->m_cyc = cc;

				printf("--> Epsilon 1\n");
				w1 = w = shortest_path(u, ugraph);
				printf("--> Epsilon 2\n");

				printf("w1:%d  w:%d  u:%d\n", w1->m_id, w->m_id, u->m_id);

				if (w != NULL)
				{
					printf("--> Epsilon 2.1.1\n");
					while (w1 != NULL)
					{
						w2 = w1;
						w1 = w1->m_father;
					}
					printf("--> Epsilon 2.1.2\n");
					v->m_cyc = 0;
					u->m_cyc = 0;

					for (o = w2->m_nghs; o != NULL; o = o->m_next)
					{
						if (o->m_v->m_cyc != 0)
						{
							break;
						}
					}
					printf("--> Epsilon 2.1.3\n");
					v->m_cyc = cc;
					u->m_cyc = cc;
					printf("--> Epsilon 2.1.4\n");
					if (cc > 1)
					{
						printf("--> Epsilon 2.1.4.0:%p\n", o);
						UVERTEX *b = o->m_v;
						printf("--> Epsilon 2.1.4.1\n");
						shortest_back_path(m->m_v, o->m_v, ugraph);
						printf("--> Epsilon 2.1.4.2\n");
						while (b != m->m_v)
						{
							printf("back:%d\n", b->m_id);
							printf("--> Epsilon 2.1.4.3\n");
							add_back_vertex_cycle(b, &cycle, cpool);
							printf("--> Epsilon 2.1.4.4\n");
							b = b->m_son;
						}
						printf("--> Epsilon 2.1.4.5\n");
						add_back_vertex_cycle(b, &cycle, cpool);
						printf("--> Epsilon 2.1.4.6\n");
						printf("back:%d\n", b->m_id);
					}
					printf("--> Epsilon 2.1.5\n");
					if (m != NULL)
					{
						add_vertex_cycle(m->m_v, &cycle, cpool);
					}
					printf("--> Epsilon 2.1.6\n");
					while (w != NULL)
					{
						add_vertex_cycle(w, &cycle, cpool);
						w->m_cyc = cc;
						w = w->m_father;
					}
					printf("--> Epsilon 2.1.7\n");
					if (o != NULL)
					{
						add_vertex_cycle(o->m_v, &cycle, cpool);
					}
					printf("--> Epsilon 2.1\n");
				}
				else
				{
					printf("--> Epsilon 2.2.1\n");
					v->m_cyc = 0;
					u->m_cyc = 0;
					printf("--> Epsilon 2.2.2\n");
					for (o = u->m_nghs; o != NULL; o = o->m_next)
					{
						if (o->m_v->m_cyc != 0)
						{
							break;
						}
					}
					printf("--> Epsilon 2.2.3\n");
					v->m_cyc = cc;
					u->m_cyc = cc;
					printf("--> Epsilon 2.2.4\n");
					if (cc > 1)
					{
						UVERTEX *b = o->m_v;
						shortest_back_path(m->m_v, o->m_v, ugraph);

						while (b != m->m_v)
						{
							add_back_vertex_cycle(b, &cycle, cpool);
							printf("Back:%d\n", b->m_id);
							b = b->m_son;
						}
						add_back_vertex_cycle(b, &cycle, cpool);
						printf("Back:%d\n", b->m_id);
					}
					printf("--> Epsilon 2.2.5\n");
					if (m != NULL)
					{
						add_vertex_cycle(m->m_v, &cycle, cpool);
					}
					printf("--> Epsilon 2.2.6\n");
					if (v != NULL)
					{
						add_vertex_cycle(v, &cycle, cpool);
					}
					printf("--> Epsilon 2.2.7\n");
					if (u != NULL)
					{
						add_vertex_cycle(u, &cycle, cpool);
					}
					printf("--> Epsilon 2.2.8\n");
					if (o != NULL)
					{
						add_vertex_cycle(o->m_v, &cycle, cpool);
					}
					printf("--> Epsilon 2.2\n");
					
				}
				printf("--> Epsilon 2.3\n");
				add_cycle_cdeco(&cycle, cdeco, cpool);
				printf("--> Epsilon 2.4\n");
			}
		}
	}
}


void rotate_cycle_left(CYCLE *cycle, CDECO *cdeco)
{
	int b;
	CVERTEX *f, *u, *v;

	f = NULL;
	u = cycle->m_root;

	while (u != NULL)
	{
		if (u->m_v->m_rob == -1)
		{
			f = u;
			b = 0;
			break;
		}
		u = u->m_next;
	}

	v = cycle->m_back_root;

	while (v != NULL)
	{
		if (v->m_v->m_rob == -1)
		{
			f = v;
			b = 1;
			break;
		}
		v = v->m_next;
	}
	
	printf("Rotating\n");

	if (f != NULL)
	{
		printf("  start rotation\n");
		if (b == 0)
		{
			printf("chi 1\n");
			v = f;

			while (v->m_next != NULL)
			{
				printf("  %d --> %d\n", v->m_next->m_v->m_id, v->m_v->m_id);
				v = v->m_next;
			}

			if (cycle->m_back_root != NULL)
			{
				printf("chi 1.1\n");
				CVERTEX *x = cycle->m_back_root;
				printf("chi 1.2\n");
				v = x;
				
				while (v->m_next != NULL)
				{
					printf("  %d --> %d\n", v->m_next->m_v->m_id, v->m_v->m_id);
					v = v->m_next;
				}
			}

			printf("chi 1.7\n");

			CVERTEX *x = cycle->m_root;
			printf("chi 1.8\n");

			if (cycle->m_back_root == NULL)
			{
				printf("  %d --> %d\n", x->m_v->m_id, v->m_v->m_id);
			}

			printf("chi 1.9\n");
			v = x;
				
			while (v != f)
			{
				printf("  %d --> %d\n", v->m_next->m_v->m_id, v->m_v->m_id);
				v = v->m_next;
			}
		}
		else
		{
			while (v->m_next != NULL)
			{
				printf("  %d --> %d\n", v->m_next->m_v->m_id, v->m_v->m_id);
				v = v->m_next;
			}

			if (cycle->m_root != NULL)
			{
				printf("chi 1.1\n");
				CVERTEX *x = cycle->m_root;
				printf("chi 1.2\n");
				v = x;
				
				while (v->m_next != NULL)
				{
					printf("  %d --> %d\n", v->m_next->m_v->m_id, v->m_v->m_id);
					v = v->m_next;
				}
			}			
			printf("chi 1.7\n");

			CVERTEX *x = cycle->m_back_root;
			printf("chi 1.8\n");
			if (cycle->m_root == NULL)
			{
				printf("  %d --> %d\n", x->m_v->m_id, v->m_v->m_id);
			}

			printf("chi 1.9\n");
			v = x;
				
			while (v != f)
			{
				printf("  %d --> %d\n", v->m_next->m_v->m_id, v->m_v->m_id);
				v = v->m_next;
			}
		}
	}
}


void rotate_cycle_right(CYCLE *cycle, CDECO *cdeco)
{
	int b;
	CVERTEX *f, *u, *v;

	f = NULL;
	u = cycle->m_root;

	while (u != NULL)
	{
		if (u->m_v->m_rob == -1)
		{
			f = u;
			b = 0;
			break;
		}
		u = u->m_next;
	}

	v = cycle->m_back_root;

	while (v != NULL)
	{
		if (v->m_v->m_rob == -1)
		{
			f = v;
			b = 1;
			break;
		}
		v = v->m_next;
	}
	
	printf("Rotating\n");

	if (f != NULL)
	{
		printf("  start rotation\n");
		if (b == 0)
		{
			printf("chi 1\n");
			v = f;

			while (v != cycle->m_root)
			{
				printf("  %d --> %d\n", v->m_prev->m_v->m_id, v->m_v->m_id);
				v = v->m_prev;
			}

			if (cycle->m_back_root != NULL)
			{
				printf("chi 1.1\n");
				CVERTEX *x = cycle->m_back_root;
				printf("chi 1.2\n");
				while (x->m_next != NULL)
				{
					x = x->m_next;
				}
				printf("chi 1.2\n");
				v = x;
				
				while (v != cycle->m_back_root)
				{
					printf("  %d --> %d\n", v->m_prev->m_v->m_id, v->m_v->m_id);
					v = v->m_prev;
				}
			}

			printf("chi 1.7\n");

			CVERTEX *x = cycle->m_root;
			printf("chi 1.8\n");
			while (x->m_next != NULL)
			{
				x = x->m_next;
			}
			if (cycle->m_back_root == NULL)
			{
				printf("  %d --> %d\n", x->m_v->m_id, v->m_v->m_id);
			}

			printf("chi 1.9\n");
			v = x;
				
			while (v != f)
			{
				printf("  %d --> %d\n", v->m_prev->m_v->m_id, v->m_v->m_id);
				v = v->m_prev;
			}
		}
		else
		{
			while (v != cycle->m_back_root)
			{
				printf("  %d --> %d\n", v->m_prev->m_v->m_id, v->m_v->m_id);
				v = v->m_prev;
			}

			if (cycle->m_root != NULL)
			{
				printf("chi 1.1\n");
				CVERTEX *x = cycle->m_root;
				printf("chi 1.2\n");
				while (x->m_next != NULL)
				{
					x = x->m_next;
				}
				printf("chi 1.2\n");
				v = x;
				
				while (v != cycle->m_root)
				{
					printf("  %d --> %d\n", v->m_prev->m_v->m_id, v->m_v->m_id);
					v = v->m_prev;
				}
			}			
			printf("chi 1.7\n");

			CVERTEX *x = cycle->m_back_root;
			printf("chi 1.8\n");
			while (x->m_next != NULL)
			{
				x = x->m_next;
			}
			if (cycle->m_root == NULL)
			{
				printf("  %d --> %d\n", x->m_v->m_id, v->m_v->m_id);
			}

			printf("chi 1.9\n");
			v = x;
				
			while (v != f)
			{
				printf("  %d --> %d\n", v->m_prev->m_v->m_id, v->m_v->m_id);
				v = v->m_prev;
			}
		}
	}
}


void decompose_ucycle(UGRAPH *ugraph, CDECO *cdeco, CPOOL *cpool)
{
	UVERTEX *v;

//	init_ustack(ustack, ugraph->m_E);
//	init_ggraph(tree, gpool);

	for (v = ugraph->m_root; v != NULL; v = v->m_next)
	{
		v->m_low = -1;
		v->m_art = 0;
		v->m_cyc = 0;
		v->m_lowv = NULL;
		v->m_son = NULL;
	}
/*
	for (v = ugraph->m_root; v != NULL; v = v->m_next)
	{
		if (v->m_low == -1)
		{
			v->m_father = NULL;
			cc = 0;
			BICON_CNT = 0;
			search_ucycle_root(v, ugraph, cdeco, cpool);
		}
	}
*/
	for (v = ugraph->m_root; v != NULL; v = v->m_next)
	{
		if (v->m_cyc == 0)
		{
			++cc;

			CYCLE cycle;
			init_cycle(&cycle, cpool);

			UVERTEX *vr = v;
//			UVERTEX *vq = v->m_lowv;

			UVERTEX *last = NULL;
			UVERTEX *first = NULL;

			while (vr != NULL && vr->m_cyc == 0)
			{
				add_vertex_cycle(vr, &cycle, cpool);
				vr->m_cyc = cc;
				first = vr;
				vr = vr->m_father;
			}

/*
			if (vr->m_cyc == 0)
			{
				if (first == NULL)
				{
					for (n = vr->m_nghs; n != NULL; n = n->m_next)
					{
						if (n->m_v->m_cyc != 0 && n->m_v->m_cyc != cc)
						{
							add_vertex_cycle(n->m_v, &cycle, cpool);
							break;
						}
					}
				}
				printf("%d x:%d\n", vr->m_id, vr->m_father->m_id);
				add_vertex_cycle(vr, &cycle, cpool);
				first = vr;
				vr->m_cyc = cc;
				fin = vr->m_father;
				last = vr;
			}
*/
//			vr->m_cyc = 1;
//			vr->m_father->m_cyc = 1;
//			vr = vr->m_father;
//		}
/*
			while (vq->m_cyc == 0)
			{
				printf("chi 1\n");
				printf("chi 2:%d\n", vq->m_id);
				add_vertex_cycle(vq, &cycle, cpool);
				printf("chi 3\n");
				vq->m_cyc = cc;
				printf("chi 4\n");
				first = vq;
				vq = vq->m_son;
				printf("chi 5\n");
			}
*/
//		vr->m_cyc = 1;
//		vr->m_lowv->m_cyc = 1;
/*
		while (vr->m_father != NULL)
		{
			if (vr->m_cyc == 0)
			{
				if (first == NULL)
				{
					for (n = vr->m_nghs; n != NULL; n = n->m_next)
					{
						printf("cccc:%d,%d,%d\n", n->m_v->m_id, n->m_v->m_cyc, cc);
						if (n->m_v->m_cyc != 0 && n->m_v->m_cyc != cc)
						{
							add_vertex_cycle(n->m_v, &cycle, cpool);
							break;
						}
					}
				}
				printf("%d x:%d\n", vr->m_id, vr->m_father->m_id);
				add_vertex_cycle(vr, &cycle, cpool);
				first = vr;
				vr->m_cyc = cc;
				fin = vr->m_father;
				last = vr;
			}
*/
//			vr->m_cyc = 1;
//			vr->m_father->m_cyc = 1;
//			vr = vr->m_father;
//		}
//		printf("xx:%d\n", vr->m_id);
/*
		if (cdeco->m_size == 0)
		{
			if (fin != NULL)
			{
				fin->m_cyc = cc;
				add_vertex_cycle(fin, &cycle, cpool);
				last = fin;
			}
		}
		if (last != NULL)
		{
			if (cc > 1)
			{
				for (n = last->m_nghs; n != NULL; n = n->m_next)
				{
					printf("ccbb:%d,%d,%d,%d\n", last->m_id, n->m_v->m_id, n->m_v->m_cyc, cc);
					if (n->m_v->m_cyc != 0 && n->m_v->m_cyc != cc)
					{
						add_vertex_cycle(n->m_v, &cycle, cpool);
						break;
					}
				}
			}
		}
		print_cycle(&cycle);
*/
			add_cycle_cdeco(&cycle, cdeco, cpool);
		}
	}
//	destroy_stack();
}


void init_stack(GRAPH *graph)
{
	STACK_edges = (STACK_EDGE*)malloc(graph->m_E * sizeof(STACK_EDGE));
	assert(STACK_edges != NULL);

	STACK_last_edge = STACK_edges;
}


void destroy_stack(void)
{
	free(STACK_edges);
}


void search_biconnected(GRAPH *graph, VERTEX *v)
{
	v->m_val = BICON_CNT++;
	v->m_low = v->m_val;

	VERTEX **n;
	
	for (n = v->m_nghs; *n != NULL; ++n)
	{
		STACK_last_edge->m_v1 = v;
		STACK_last_edge->m_v2 = (*n);
		++STACK_last_edge;

		printf("push:%d,%d\n", v - graph->m_vrtxs, (*n) - graph->m_vrtxs);

		if ((*n)->m_low == -1)
		{
			(*n)->m_f = v;
			search_biconnected(graph, *n);

			if ((*n)->m_low >= v->m_val && v->m_val != 0)
			{
				printf("--> arti:%d <--\n", v - graph->m_vrtxs);
				print_graph(graph);

				while (STACK_last_edge > STACK_edges)
				{
					--STACK_last_edge;

					if (v == STACK_last_edge->m_v1 && (*n) == STACK_last_edge->m_v2)
					{
						break;
					}
					if (STACK_last_edge->m_v1->m_2cc == -1)
					{
						if (STACK_last_edge->m_v2->m_2cc == -1)
						{
							STACK_last_edge->m_v1->m_2cc = COMP2_CNT;
							STACK_last_edge->m_v2->m_2cc = COMP2_CNT;
						}
						else
						{
							if (STACK_last_edge->m_v2->m_2cc != COMP2_CNT)
							{
								printf("** Bridge:%d,%d\n", STACK_last_edge->m_v1 - graph->m_vrtxs, STACK_last_edge->m_v2 - graph->m_vrtxs);
							}
							STACK_last_edge->m_v1->m_2cc = COMP2_CNT;
						}
					}
					else
					{
						if (STACK_last_edge->m_v2->m_2cc == -1)
						{
							if (STACK_last_edge->m_v1->m_2cc != COMP2_CNT)
							{
								printf("** Bridge:%d,%d\n", STACK_last_edge->m_v1 - graph->m_vrtxs, STACK_last_edge->m_v2 - graph->m_vrtxs);
							}
							STACK_last_edge->m_v2->m_2cc = COMP2_CNT;
						}
					}
				}
				++COMP2_CNT;

				printf("pop:%d,%d\n", v - graph->m_vrtxs, (*n) - graph->m_vrtxs);
			}

			v->m_low = (*n)->m_low < v->m_low ? (*n)->m_low : v->m_low;
		}
		else
		{
			if ((*n) != v->m_f)
			{
				v->m_low = (*n)->m_val < v->m_low ? (*n)->m_val : v->m_low;
			}
		}
	}
}


void search_biconnected_root(GRAPH *graph, VERTEX *v)
{
	v->m_val = BICON_CNT++;
	v->m_low = v->m_val;

	VERTEX **n;

	int desc = 0;
	
	for (n = v->m_nghs; *n != NULL; ++n)
	{
		STACK_last_edge->m_v1 = v;
		STACK_last_edge->m_v2 = (*n);
		++STACK_last_edge;

		printf("push:%d,%d\n", v - graph->m_vrtxs, (*n) - graph->m_vrtxs);

		if ((*n)->m_low == -1)
		{
			++desc;

			(*n)->m_f = v;
			search_biconnected(graph, *n);

			if ((*n)->m_low >= v->m_val && v->m_val != 0)
			{
				printf("--> arti:%d <--\n", v - graph->m_vrtxs);

				while (STACK_last_edge > STACK_edges)
				{
					--STACK_last_edge;

					if (v == STACK_last_edge->m_v1 && (*n) == STACK_last_edge->m_v2)
					{
						break;
					}

					if (STACK_last_edge->m_v1->m_2cc == -1)
					{
						STACK_last_edge->m_v1->m_2cc = COMP2_CNT;
					}
					if (STACK_last_edge->m_v2->m_2cc == -1)
					{
						STACK_last_edge->m_v2->m_2cc = COMP2_CNT;
					}
				}
				++COMP2_CNT;

				printf("pop:%d,%d\n", v - graph->m_vrtxs, (*n) - graph->m_vrtxs);
			}

			v->m_low = (*n)->m_low < v->m_low ? (*n)->m_low : v->m_low;
		}
		else
		{
			if ((*n) != v->m_f)
			{
				v->m_low = (*n)->m_val < v->m_low ? (*n)->m_val : v->m_low;
			}
		}
	}
	printf("root:%d\n", desc);

//	--COMP2_CNT;

	while (STACK_last_edge > STACK_edges)
	{
		--STACK_last_edge;
		
		if (STACK_last_edge->m_v1->m_2cc == -1)
		{
			if (STACK_last_edge->m_v2->m_2cc == -1)
			{
				STACK_last_edge->m_v1->m_2cc = COMP2_CNT;
				STACK_last_edge->m_v2->m_2cc = COMP2_CNT;
			}
			else
			{
				if (STACK_last_edge->m_v2->m_2cc != COMP2_CNT)
				{
					printf("** Bridge:%d,%d\n", STACK_last_edge->m_v1 - graph->m_vrtxs, STACK_last_edge->m_v2 - graph->m_vrtxs);
				}
				STACK_last_edge->m_v1->m_2cc = COMP2_CNT;
			}
		}
		else
		{
			if (STACK_last_edge->m_v2->m_2cc == -1)
			{
				if (STACK_last_edge->m_v1->m_2cc != COMP2_CNT)
				{
					printf("** Bridge:%d,%d\n", STACK_last_edge->m_v1 - graph->m_vrtxs, STACK_last_edge->m_v2 - graph->m_vrtxs);
				}
				STACK_last_edge->m_v2->m_2cc = COMP2_CNT;
			}
		}
		
		printf("r:%d,%d\n", STACK_last_edge->m_v1 - graph->m_vrtxs, STACK_last_edge->m_v2 - graph->m_vrtxs);
	}
}
 

void decompose_biconnected(GRAPH *graph)
{
	unsigned int v;

	init_stack(graph);

	for (v = 0; v < graph->m_V; ++v)
	{
		graph->m_vrtxs[v].m_low = -1;
		graph->m_vrtxs[v].m_2cc = -1;
	}

	for (v = 0; v < graph->m_V; ++v)
	{
		if (graph->m_vrtxs[v].m_low == -1)
		{
			graph->m_vrtxs[v].m_f = NULL;
			search_biconnected_root(graph, graph->m_vrtxs + v);
		}
	}

	destroy_stack();
}

      
void init_star(MAPPING *start, MAPPING *goal, GRAPH *graph)
{
	SOL_last_mv = solution;

	STATE_vls = (int*)malloc(STAR_QUEUE * graph->m_V * sizeof(int));
	assert(STATE_vls != NULL);
/*
	CLOSE_vls = (int*)malloc(CLOSE_LIST * goal->m_R * sizeof(int));
	assert(CLOSE_vls != NULL);

	CLOSE_last_vl = CLOSE_vls;
	
	OPEN_vls = (int*)malloc(OPEN_LIST * goal->m_R * sizeof(int));
	assert(OPEN_vls != NULL);

	OPEN_last_vl = OPEN_vls;
*/
	CLOSE_nds = (LIST_NODE*)malloc(CLOSE_LIST * sizeof(LIST_NODE));
	assert(CLOSE_nds != NULL);

	CLOSE_last_nd = CLOSE_nds;
	CLOSE_root = CLOSE_last_nd++;
	CLOSE_root->m_nds = NULL;

	CLOSE_next_nds = (LIST_NODE**)malloc(CLOSE_LIST * graph->m_V * sizeof(LIST_NODE*));
	assert(CLOSE_next_nds != NULL);

	CLOSE_last_nnd = CLOSE_next_nds;

	OPEN_nds = (LIST_NODE*)malloc(OPEN_LIST * sizeof(LIST_NODE));
	assert(OPEN_nds != NULL);

	OPEN_last_nd = OPEN_nds;
	OPEN_root = OPEN_last_nd++;
	OPEN_root->m_nds = NULL;

	OPEN_next_nds = (LIST_NODE**)malloc(OPEN_LIST * graph->m_V * sizeof(LIST_NODE*));
	assert(OPEN_next_nds != NULL);

	OPEN_last_nnd = OPEN_next_nds;

	STATE_vrtxs = (VERTEX**)malloc(STAR_QUEUE * goal->m_R * sizeof(VERTEX*));
	assert(STATE_vrtxs != NULL);

	STATE_last_vl = STATE_vls;
	STATE_last_vrtx = STATE_vrtxs;

	STATE_sts = (STATE*)malloc(STAR_QUEUE * sizeof(STATE));
	assert(STATE_sts != NULL);

	STATE_sts_heap = (STATE**)malloc(STAR_QUEUE * sizeof(STATE*));
	assert(STATE_sts_heap != NULL);

	unsigned int i;

	for (i = 0; i < STAR_QUEUE; ++i)
	{
		STATE_sts_heap[i] = STATE_sts + i;
		STATE_sts[i].m_vls = STATE_vls + i * graph->m_V;
		STATE_sts[i].m_vrtxs = STATE_vrtxs + i * goal->m_R;
	}

	STATE_last_st = STATE_sts_heap;

	setup_distances(graph, &dist_B);
	all_pairs_shortest(&dist_B);

	for (i = 0; i < graph->m_V; ++i)
	{
		(*STATE_last_st)->m_vls[i] = -1;
	}
	for (i = 0; i < start->m_R; ++i)
	{
		(*STATE_last_st)->m_vls[start->m_vrtxs[i] - graph->m_vrtxs] = i;
		(*STATE_last_st)->m_vrtxs[i] = start->m_vrtxs[i];
	}
	(*STATE_last_st)->m_q = get_quality_star((*STATE_last_st), goal, graph);
	(*STATE_last_st)->m_d = 0;
	(*STATE_last_st)->m_mv = SOL_last_mv++;

	(*STATE_last_st)->m_mv->m_prev = NULL;
	(*STATE_last_st)->m_mv->m_from = NULL;
	(*STATE_last_st)->m_mv->m_to = NULL;

	print_state(*STATE_last_st, start, graph);
	++STATE_last_st;

	TEMP_state.m_vrtxs = (VERTEX**)malloc(graph->m_V * sizeof(VERTEX*));
	assert(TEMP_state.m_vrtxs != NULL);

	TEMP_state.m_vls = (int*)malloc(goal->m_R * sizeof(int));
	assert(TEMP_state.m_vls != NULL);
}


void print_state(STATE *state, MAPPING *start, GRAPH *graph)
{
	unsigned int i;

	printf("Quality: %d\n", state->m_q); 

	printf("Graph: ");
	for (i = 0; i < graph->m_V; ++i)
	{
		printf("%d ", state->m_vls[i]);
	}
	printf("\n");

	printf("Positions: ");
	for (i = 0; i < start->m_R; ++i)
	{
		printf("%d ", state->m_vrtxs[i] - graph->m_vrtxs);
	}
	printf("\n");
}


void print_heap(STATE **heap, STATE **last_state, MAPPING *start, GRAPH *graph)
{
	printf("Heap\n");

	while (heap <= last_state)
	{
		print_state(*heap, start, graph);
		++heap;
	}
}


int get_quality_star(STATE *state, MAPPING *goal, GRAPH *graph)
{
	unsigned int r, q = 0;

	for (r = 0; r < goal->m_R; ++r)
	{
		unsigned int u = state->m_vrtxs[r] - graph->m_vrtxs;
		unsigned int v = goal->m_vrtxs[r] - graph->m_vrtxs;

		q += dist_B.m_d[u][v];
//		printf("%d (%d,%d)  ", dist_B.m_d[u][v], u, v);
	}
//	printf("\n");

	return q;
}


void print_solution(MOVE *m, GRAPH *graph, int d)
{
	if (m->m_prev != NULL)
	{
		print_solution(m->m_prev, graph, d+1);
		printf("%d:%d->%d ", d, m->m_from - graph->m_vrtxs, m->m_to - graph->m_vrtxs);
	}
}


int search_star(MAPPING *goal, GRAPH *graph)
{
	while(1)
	{
		++STAT_steps;

		--STATE_last_st;

//		print_solution((*STATE_sts_heap)->m_mv, graph);
//		printf("\n\n");

		if (STATE_last_st < STATE_sts_heap)
		{
			printf("Unsolvable !\n");
			return -1;
		}
		if ((*STATE_sts_heap)->m_q == 0)
		{
			printf("Solved !\n");
			print_solution((*STATE_sts_heap)->m_mv, graph, 0);
			printf("\n");
			return 1;
		}

//		printf("Close !\n");
		close_state_tree(*STATE_sts_heap, goal, graph);
//		close_state(*STATE_sts_heap, goal, graph);
//		print_state(*STATE_sts_heap, goal, graph);
//		print_closed(goal);

		STATE *state = *STATE_sts_heap;

//		printf("Steps:%llu Q:%d D:%d C:%d O:%d\n", STAT_steps, state->m_q, state->m_d, CLOSE_last_vl - CLOSE_vls, OPEN_last_vl - OPEN_vls);

		STATE *temp;
		temp = *STATE_sts_heap;
		*STATE_sts_heap = *STATE_last_st;
		*STATE_last_st = temp;

//		printf("Down:%d\n\n", STATE_last_st - STATE_sts_heap);
//		print_heap(STATE_sts_heap, STATE_last_st, goal, graph);

		if (STATE_last_st - 1 >= STATE_sts_heap)
		{
			down_heap(STATE_sts_heap, STATE_last_st - 1);
		}

//		printf("Alpha 1\n");
//		print_heap(STATE_sts_heap, STATE_last_st, goal, graph);

		unsigned int r;
		unsigned int i;
						
		for (i = 0; i < goal->m_R; ++i)
		{
			TEMP_state.m_vrtxs[i] = state->m_vrtxs[i];
		}
		for (i = 0; i < graph->m_V; ++i)
		{
			TEMP_state.m_vls[i] = state->m_vls[i];
		}
		TEMP_state.m_q = state->m_q;
		TEMP_state.m_d = state->m_d;
		TEMP_state.m_mv = state->m_mv;

		for (r = 0; r < goal->m_R; ++r)
		{
			VERTEX **n = TEMP_state.m_vrtxs[r]->m_nghs;
			
			while (*n != NULL)
			{
				unsigned int hs = 0;
				unsigned int ni = *n - graph->m_vrtxs;
				
				if (TEMP_state.m_vls[ni] == -1)
				{
					for (i = 0; i < goal->m_R; ++i)
					{
						(*STATE_last_st)->m_vrtxs[i] = TEMP_state.m_vrtxs[i];
					}
					for (i = 0; i < graph->m_V; ++i)
					{
						(*STATE_last_st)->m_vls[i] = TEMP_state.m_vls[i];
					}

					unsigned int nr = TEMP_state.m_vrtxs[r] - graph->m_vrtxs;
					VERTEX *from = (*STATE_last_st)->m_vrtxs[r];

					(*STATE_last_st)->m_vls[nr] = -1;
					(*STATE_last_st)->m_vls[ni] = r;
					(*STATE_last_st)->m_vrtxs[r] = *n;

					unsigned int v = goal->m_vrtxs[r] - graph->m_vrtxs;

					(*STATE_last_st)->m_q = TEMP_state.m_q - dist_B.m_d[nr][v] + dist_B.m_d[ni][v];
					(*STATE_last_st)->m_d = TEMP_state.m_d + 1;

					if (!is_closed_tree(*STATE_last_st, goal, graph) && !is_open_tree(*STATE_last_st, goal, graph))
					{
						MOVE *m = SOL_last_mv++;
						
						m->m_prev = TEMP_state.m_mv;
						m->m_from = from;
						m->m_to = *n;

						(*STATE_last_st)->m_mv = m;

						open_state_tree(*STATE_last_st, goal, graph);
						up_heap(STATE_sts_heap, STATE_last_st);
							
						if ((hs = STATE_last_st - STATE_sts_heap) < STAR_QUEUE)
						{
							++STATE_last_st;
						}
						else
						{
							printf("Out of heap (A) !\n");
							printf("Unable to solve !\n");
							exit(-3);
						}
					}
				}
				if (hs >= STAR_QUEUE)
				{
					printf("Out of heap (B) !\n");
					printf("Unable to solve !\n");
					exit(-4);
				}
				++n;
			}
		}
	}

	return -1;
}


void up_heap(STATE **heap, STATE **last_state)
{
	STATE *temp;

	unsigned int i = last_state - heap;
	unsigned int j = (i + 1) / 2 - 1;

	while (i > 0 && heap[i]->m_q < heap[j]->m_q)
	{
		temp = heap[i];
		heap[i] = heap[j];
		heap[j] = temp;

		i = j;
		j = (i + 1) / 2 - 1;
	}
}


void down_heap(STATE **heap, STATE **last_state)
{
	STATE *temp;

	unsigned int m = last_state - heap;
	unsigned int i = 0;
	unsigned int j, k;

	while (1)
	{
		j = 2 * i + 1;
		k = 2 * i + 2;

		if (k <= m)
		{
			if (heap[j]->m_q > heap[k]->m_q)
			{
				if (heap[k]->m_q < heap[i]->m_q)
				{
					temp = heap[i];
					heap[i] = heap[k];
					heap[k] = temp;
					i = k;
				}
				else
				{
					break;
				}
			}
			else
			{
				if (heap[j]->m_q < heap[i]->m_q)
				{
					temp = heap[i];
					heap[i] = heap[j];
					heap[j] = temp;
					i = j;
				}
				else
				{
					break;
				}
			}
		}
		else
		{
			if (j <= m)
			{
				if (heap[j]->m_q < heap[i]->m_q)
				{
					temp = heap[i];
					heap[i] = heap[j];
					heap[j] = temp;
					i = j;
					break;
				}
				else
				{
					break;
				}
			}
			else
			{
				break;
			}
		}
	}
}


void close_state_tree(STATE *state, MAPPING *goal, GRAPH *graph)
{
	LIST_NODE *node = CLOSE_root;

	unsigned int r, i;

	for (r = 0; r < goal->m_R; ++r)
	{
		if (node->m_nds != NULL)
		{
			node = node->m_nds[state->m_vrtxs[r] - graph->m_vrtxs];
		}
		else
		{
			break;
		}
	}

	while (r < goal->m_R)
	{
		node->m_nds = CLOSE_last_nnd;
		CLOSE_last_nnd += graph->m_V;

		for (i = 0; i < graph->m_V; ++i)
		{
			node->m_nds[i] = CLOSE_last_nd++;
			node->m_nds[i]->m_nds = NULL;
		}
		node = node->m_nds[state->m_vrtxs[r] - graph->m_vrtxs];
		++r;
	}
	node->m_nds = (LIST_NODE**)1;

	if (CLOSE_last_nd - CLOSE_nds >= CLOSE_LIST)
	{
		printf("Close list full !\n");
		printf("Unable to solve !\n");
		exit(-1);
	}
}


int is_closed_tree(STATE *state, MAPPING *goal, GRAPH *graph)
{
	LIST_NODE *node = CLOSE_root;

	unsigned int r;

	for (r = 0; r < goal->m_R; ++r)
	{
		if (node->m_nds != NULL)
		{
			node = node->m_nds[state->m_vrtxs[r] - graph->m_vrtxs];
		}
		else
		{
			break;
		}
	}

	if (r == goal->m_R && node->m_nds != NULL)
	{
		return 1;
	}
	
	return 0;
}


void open_state_tree(STATE *state, MAPPING *goal, GRAPH *graph)
{
	LIST_NODE *node = OPEN_root;

	unsigned int r, i;

	for (r = 0; r < goal->m_R; ++r)
	{
		if (node->m_nds != NULL)
		{
			node = node->m_nds[state->m_vrtxs[r] - graph->m_vrtxs];
		}
		else
		{
			break;
		}
	}

	while (r < goal->m_R)
	{
		node->m_nds = OPEN_last_nnd;
		OPEN_last_nnd += graph->m_V;

		for (i = 0; i < graph->m_V; ++i)
		{
			node->m_nds[i] = OPEN_last_nd++;
			node->m_nds[i]->m_nds = NULL;
		}
		node = node->m_nds[state->m_vrtxs[r] - graph->m_vrtxs];
		++r;
	}
	node->m_nds = (LIST_NODE**)1;

	if (OPEN_last_nd - OPEN_nds >= OPEN_LIST)
	{
		printf("Open list full !\n");
		printf("Unable to solve !\n");
		exit(-2);
	}
}


int is_open_tree(STATE *state, MAPPING *goal, GRAPH *graph)
{
	LIST_NODE *node = OPEN_root;

	unsigned int r;

	for (r = 0; r < goal->m_R; ++r)
	{
		if (node->m_nds != NULL)
		{
			node = node->m_nds[state->m_vrtxs[r] - graph->m_vrtxs];
		}
		else
		{
			break;
		}
	}

	if (r == goal->m_R && node->m_nds != NULL)
	{
		return 1;
	}
	
	return 0;
}


void print_closed_tree(LIST_NODE *node, MAPPING *goal)
{
	unsigned int r;

	if (node->m_nds != NULL)
	{
		for (r = 0; r < goal->m_R; ++r)
		{
			printf("%d ", r);
			print_closed_tree(node->m_nds[r], goal);
		}
	}
	else
	{
		printf("-\n");
	}
}


void close_state(STATE *state, MAPPING *goal, GRAPH *graph)
{
	unsigned int r;

	for (r = 0; r < goal->m_R; ++r)
	{
		*CLOSE_last_vl++ = state->m_vrtxs[r] - graph->m_vrtxs;
	}
	if (CLOSE_last_vl - CLOSE_vls >= (int)goal->m_R * CLOSE_LIST)
	{
		printf("Close list full !\n");
	}
}


int is_closed(STATE *state, MAPPING *goal, GRAPH *graph)
{
	unsigned int r;
	int *v = CLOSE_vls;

	while (v < CLOSE_last_vl)
	{
		int *vl_c = v;

		for (r = 0; r < goal->m_R; ++r)
		{
			if (state->m_vrtxs[r] - graph->m_vrtxs != *vl_c)
			{
				break;
			}
			++vl_c;
		}
		if (r == (int)goal->m_R)
		{
			return 1;
		}
		v += goal->m_R;
	}

	return 0;
}


void print_closed(MAPPING *goal)
{
	unsigned int r;
	int *v = CLOSE_vls;

	printf("Closed list\n");

	while (v < CLOSE_last_vl)
	{
		int *vl_c = v;

		for (r = 0; r < goal->m_R; ++r)
		{
			printf("%d ", *vl_c);
			++vl_c;
		}
		printf("\n");
		v += goal->m_R;
	}
}


void open_state(STATE *state, MAPPING *goal, GRAPH *graph)
{
	unsigned int r;

	for (r = 0; r < goal->m_R; ++r)
	{
		*OPEN_last_vl++ = state->m_vrtxs[r] - graph->m_vrtxs;
	}
	if (OPEN_last_vl - OPEN_vls >= (int)goal->m_R * OPEN_LIST)
	{
		printf("Open list full !\n");
	}
}


int is_open(STATE *state, MAPPING *goal, GRAPH *graph)
{
	unsigned int r;
	int *v = OPEN_vls;

	while (v < OPEN_last_vl)
	{
		int *vl_c = v;

		for (r = 0; r < goal->m_R; ++r)
		{
			if (state->m_vrtxs[r] - graph->m_vrtxs != *vl_c)
			{
				break;
			}
			++vl_c;
		}
		if (vl_c - v == (int)goal->m_R)
		{
			return 1;
		}
		v += goal->m_R;
	}

	return 0;
}


void print_open(MAPPING *goal)
{
	unsigned int r;
	int *v = OPEN_vls;

	printf("Open list\n");

	while (v < OPEN_last_vl)
	{
		int *vl_c = v;

		for (r = 0; r < goal->m_R; ++r)
		{
			printf("%d ", *vl_c);
			++vl_c;
		}
		printf("\n");
		v += goal->m_R;
	}
}


void setup_distances(GRAPH *graph, DIST *dist)
{
	unsigned int v, u;

	dist->m_V = graph->m_V;

	dist->m_d = (int**)malloc(graph->m_V * sizeof(int*));
	assert(dist->m_d != NULL);

	for (v = 0; v < graph->m_V; ++v)
	{
		dist->m_d[v] = (int*)malloc(graph->m_V * sizeof(int));
		assert(dist->m_d[v]);

		for (u = 0; u < graph->m_V; ++u)
		{
			dist->m_d[v][u] = graph->m_V + 1;
		}
		dist->m_d[v][v] = 0;
	}

	for (v = 0; v < graph->m_V; ++v)
	{
		VERTEX **n = graph->m_vrtxs[v].m_nghs;

		while (*n != NULL)
		{
			u = *n - graph->m_vrtxs;
			dist->m_d[v][u] = 1;
			++n;
		}
	}
}


void print_distances(DIST *dist)
{
	unsigned int v, u;

	for (v = 0; v < dist->m_V; ++v)
	{
		for (u = 0; u < dist->m_V; ++u)
		{
			printf("%d\t", dist->m_d[v][u]);
		}
		printf("\n");
	}
}


void multiply_distances(DIST *dist)
{
	unsigned int v, u, x;

	for (v = 0; v < dist->m_V; ++v)
	{
		for (u = 0; u < dist->m_V; ++u)
		{
			for (x = 0; x < dist->m_V; ++x)
			{
				int nd = dist->m_d[v][x] + dist->m_d[x][u];
				dist->m_d[v][u] = (nd < dist->m_d[v][u]) ? nd : dist->m_d[v][u];
			}
		}
	}
}


void all_pairs_shortest(DIST *dist)
{
	unsigned int i, l = ceil(log(dist->m_V)/log(2));

	printf("le:%d\n", l);

	for (i = 0; i <= l; ++i)
	{
		multiply_distances(dist);
	}
}


void destroy_distances(DIST *dist)
{
	unsigned int v;

	for (v = 0; v < dist->m_V;++v)
	{
		free(dist->m_d[v]);
	}

	free(dist->m_d);
}


void graph2string(GRAPH *graph, char *output)
{
	sprintf(output, "p edge %u %u\n", graph->m_V, graph->m_E);
	output += strlen(output);

	VERTEX *v, *t, **n;

	for (v = graph->m_vrtxs, t = graph->m_vrtxs + graph->m_V; v < t; ++v)
	{

		for (n = v->m_nghs; *n != NULL; ++n)
		{
			sprintf(output, "e %u %u\n", v - graph->m_vrtxs, *n - graph->m_vrtxs);
			output += strlen(output);
		}
	} 
}


void print_graph(GRAPH *graph)
{
	printf("p edge %u %u\n", graph->m_V, graph->m_E);

	VERTEX *v, *t, **n;

	for (v = graph->m_vrtxs, t = graph->m_vrtxs + graph->m_V; v < t; ++v)
	{
		printf("v:%d dfs:%d low:%d 2cc:%d\n", v - graph->m_vrtxs, v->m_val, v->m_low, v->m_2cc);
	}

	for (v = graph->m_vrtxs, t = graph->m_vrtxs + graph->m_V; v < t; ++v)
	{
		for (n = v->m_nghs; *n != NULL; ++n)
		{
			printf("e %u %u\n", v - graph->m_vrtxs, *n - graph->m_vrtxs);
		}
	} 
}


void string2graph(char *input, GRAPH *graph)
{
	char *inp;
	unsigned int i, u, v;

	sscanf(input, "p edge %u %u\n", &graph->m_V, &graph->m_E);
	input = strchr(input, '\n') + 1;

	graph->m_vrtxs = (VERTEX*)malloc(graph->m_V * sizeof(VERTEX));

#ifdef CHECK_ERRORS
	if (graph->m_vrtxs == NULL)
	{
		printf("Error 8: Not enough memory.\n");
		exit(-8);
	}
#endif

	graph->m_edges = (VERTEX**)malloc((graph->m_V + graph->m_E) * sizeof(VERTEX*));

#ifdef CHECK_ERRORS
	if (graph->m_edges == NULL)
	{
		printf("Error 9: Not enough memory.\n");
		exit(-9);
	}
#endif

	unsigned int *c_E, *c, *t;

	c_E = (unsigned int*)malloc(graph->m_V * sizeof(unsigned int));

#ifdef CHECK_ERRORS
	if (c_E == NULL)
	{
		printf("Error 10: Not enough memory.\n");
		exit(-10);
	}
#endif

	for (c = c_E, t = c_E + graph->m_V; c < t; ++c)
	{
		*c = 0;
	}

	inp = input;

	for (i = 0; i < graph->m_E; ++i)
	{
		sscanf(inp, "e %u %u\n", &u, &v);
		inp = strchr(inp, '\n') + 1;
		++c_E[u];
	}

	VERTEX **ed, *vx, *tv;

	ed = graph->m_edges;
	c = c_E;

	for (vx = graph->m_vrtxs, tv = graph->m_vrtxs + graph->m_V; vx < tv; ++vx)
	{
		vx->m_nghs = ed;
		ed += *c;
		*ed = NULL;
		++ed;
		++c;
	}

	for (i = 0; i < graph->m_E; ++i)
	{
		sscanf(input, "e %u %u\n", &u, &v);
		input = strchr(input, '\n') + 1;
		*graph->m_vrtxs[u].m_nghs = graph->m_vrtxs + v;
		++graph->m_vrtxs[u].m_nghs;
	}

	c = c_E;

	for (vx = graph->m_vrtxs, tv = graph->m_vrtxs + graph->m_V; vx < tv; ++vx)
	{
		vx->m_nghs -= *c;
		++c;
	}

	free(c_E);
}


void read_graph(char *filename, GRAPH *graph)
{
	FILE *fr;

	fr = fopen(filename, "r");

#ifdef CHECK_ERRORS
	if (fr == NULL)
	{
		printf("Error 13: Cannot open file %s.\n", filename);
		exit(-13);
	}
#endif

	char *inp;
	unsigned int i, u, v;

	fscanf(fr, "p edge %u %u\n", &graph->m_V, &graph->m_E);
	graph->m_vrtxs = (VERTEX*)malloc(graph->m_V * sizeof(VERTEX));

#ifdef CHECK_ERRORS
	if (graph->m_vrtxs == NULL)
	{
		printf("Error 8: Not enough memory.\n");
		exit(-8);
	}
#endif

	graph->m_edges = (VERTEX**)malloc((graph->m_V + graph->m_E) * sizeof(VERTEX*));

#ifdef CHECK_ERRORS
	if (graph->m_edges == NULL)
	{
		printf("Error 9: Not enough memory.\n");
		exit(-9);
	}
#endif

	unsigned int *c_E, *c, *t;

	c_E = (unsigned int*)malloc(graph->m_V * sizeof(unsigned int));

#ifdef CHECK_ERRORS
	if (c_E == NULL)
	{
		printf("Error 10: Not enough memory.\n");
		exit(-10);
	}
#endif

	for (c = c_E, t = c_E + graph->m_V; c < t; ++c)
	{
		*c = 0;
	}

	long pos = ftell(fr);

	for (i = 0; i < graph->m_E; ++i)
	{
		fscanf(fr, "e %u %u\n", &u, &v);
		inp = strchr(inp, '\n') + 1;
		++c_E[u];
	}

	VERTEX **ed, *vx, *tv;

	ed = graph->m_edges;
	c = c_E;

	for (vx = graph->m_vrtxs, tv = graph->m_vrtxs + graph->m_V; vx < tv; ++vx)
	{
		vx->m_nghs = ed;
		ed += *c;
		*ed = NULL;
		++ed;
		++c;
	}

	fseek(fr, pos, SEEK_SET);

	for (i = 0; i < graph->m_E; ++i)
	{
		fscanf(fr, "e %u %u\n", &u, &v);
		*graph->m_vrtxs[u].m_nghs = graph->m_vrtxs + v;
		++graph->m_vrtxs[u].m_nghs;
	}

	c = c_E;

	for (vx = graph->m_vrtxs, tv = graph->m_vrtxs + graph->m_V; vx < tv; ++vx)
	{
		vx->m_nghs -= *c;
		++c;
	}

	free(c_E);
	fclose(fr);
}


void write_graph(char *filename, GRAPH *graph)
{
	FILE *fw;

	fw = fopen(filename, "w");

#ifdef CHECK_ERRORS
	if (fw == NULL)
	{
		printf("Error 15: Cannot open file %s for writing.\n", filename);
		exit(-15);
	}
#endif

	fprintf(fw, "p edge %u %u\n", graph->m_V, graph->m_E);

	VERTEX *v, *t, **n;

	for (v = graph->m_vrtxs, t = graph->m_vrtxs + graph->m_V; v < t; ++v)
	{

		for (n = v->m_nghs; *n != NULL; ++n)
		{
			fprintf(fw, "e %u %u\n", v - graph->m_vrtxs, *n - graph->m_vrtxs);
		}
	}

	fclose(fw);
}


void create_random_graph(unsigned int V, unsigned int E, GRAPH *graph)
{
	VERTEX *v, **l;
	unsigned int i, m, r, h, e, k;
	unsigned int *all_V, *all_E, *av, *ae, *ar, *t;

	graph->m_V = V;
	graph->m_E = E;

	graph->m_vrtxs = (VERTEX*)malloc(V * sizeof(VERTEX));

#ifdef CHECK_ERRORS
	if (graph->m_vrtxs == NULL)
	{
		printf("Error 1: Not enough memory.\n");
		exit(-1);
	}
#endif

	graph->m_edges = (VERTEX**)malloc((V + E) * sizeof(VERTEX*));

#ifdef CHECK_ERRORS
	if (graph->m_edges == NULL)
	{
		printf("Error 2: Not enough memory.\n");
		exit(-2);
	}
#endif

	all_V = (unsigned int*)malloc(V * sizeof(unsigned int));

#ifdef CHECK_ERRORS
	if (all_V == NULL)
	{
		printf("Error 3: Not enough memory.\n");
		exit(-3);
	}
#endif

	m = V * (V - 1);
	all_E = (unsigned int*)malloc(m * sizeof(unsigned int));

#ifdef CHECK_ERRORS
	if (all_E == NULL)
	{
		printf("Error 4: Not enough memory.\n");
		exit(-4);
	}
#endif

	for (av = all_V, t = all_V + V; av < t; ++av)
	{
		*av = 0;
	}

	ae = all_E;
	for (i = 0; i < m; ++i)
	{
		*ae = i;
		++ae;
	}

	ae = all_E;
	k = m;

	for (i = 0; i < E; ++i)
	{
		r = random() % k;
		ar = ae + r;

		h = *ae;
		*ae = *ar;
		*ar = h;

		++all_V[*ae % V];
		++ae;
		--k;
	}

	v = graph->m_vrtxs;
	l = graph->m_edges;

	for (av = all_V, t = all_V + V; av < t; ++av)
	{
		v->m_nghs = l;
		l += (*av + 1);
		++v;
	}

	for (ae = all_E, t = all_E + E; ae < t; ++ae)
	{
		e = *ae % V;
		h = *ae / V;

		if (e <= h)
		{
			++h;
		}

		v = graph->m_vrtxs + e;
		*(v->m_nghs) = graph->m_vrtxs + h;
		++v->m_nghs;
	}

	v = graph->m_vrtxs;

	for (av = all_V, t = all_V + V; av < t; ++av)
	{
		*v->m_nghs = NULL;
		v->m_nghs -= *av;
		++v;
	}

	free(all_V);
	free(all_E);
}


void destroy_graph(GRAPH *graph)
{
	free(graph->m_vrtxs);
	free(graph->m_edges);
}


void mapping2string(GRAPH *graph, MAPPING *map, char *output)
{
	VERTEX **v, **t;

	sprintf(output, "p robot %u\n", map->m_R);
	output += strlen(output);

	for (v = map->m_vrtxs, t = map->m_vrtxs + map->m_R; v < t; ++v)
	{
		sprintf(output, "r %u %u\n", v - map->m_vrtxs, *v - graph->m_vrtxs);
		output += strlen(output);
	} 
}


void string2mapping(char *input, GRAPH *graph, MAPPING *map)
{
	unsigned int i, x, y;

	sscanf(input, "p robot %u\n", &map->m_R);
	input = strchr(input, '\n') + 1;

	map->m_vrtxs = (VERTEX**)malloc(map->m_R * sizeof(VERTEX*));

#ifdef CHECK_ERRORS
	if (map->m_vrtxs == NULL)
	{
		printf("Error 7: Not enough memory.\n");
		exit(-7);
	}
#endif

	for (i = 0; i < map->m_R; ++i)
	{
		sscanf(input, "r %u %u\n", &x, &y);
		input = strchr(input, '\n') + 1;
		map->m_vrtxs[x] = graph->m_vrtxs + y;
	}
}


void read_mapping(char *filename, GRAPH *graph, MAPPING *map)
{
	FILE *fr;

	fr = fopen(filename, "r");

#ifdef CHECK_ERRORS
	if (fr == NULL)
	{
		printf("Error 12: Cannot open file %s.\n", filename);
		exit(-12);
	}
#endif

	unsigned int i, x, y;

	fscanf(fr, "p robot %u\n", &map->m_R);

	map->m_vrtxs = (VERTEX**)malloc(map->m_R * sizeof(VERTEX*));

#ifdef CHECK_ERRORS
	if (map->m_vrtxs == NULL)
	{
		printf("Error 7: Not enough memory.\n");
		exit(-7);
	}
#endif

	for (i = 0; i < map->m_R; ++i)
	{
		fscanf(fr, "r %u %u\n", &x, &y);
		map->m_vrtxs[x] = graph->m_vrtxs + y;
	}

	fclose(fr);
}



void write_mapping(char *filename, GRAPH *graph, MAPPING *map)
{
	FILE *fw;

	fw = fopen(filename, "w");

#ifdef CHECK_ERRORS
	if (fw == NULL)
	{
		printf("Error 16: Cannot open file %s for writing.\n", filename);
		exit(-16);
	}
#endif

	VERTEX **v, **t;

	fprintf(fw, "p robot %u\n", map->m_R);

	for (v = map->m_vrtxs, t = map->m_vrtxs + map->m_R; v < t; ++v)
	{
		fprintf(fw, "r %u %u\n", v - map->m_vrtxs, *v - graph->m_vrtxs);
	}

	fclose(fw);
}


void create_random_mapping(unsigned int R, GRAPH *graph, MAPPING *map)
{
	unsigned int i, k, r;
	VERTEX *v, *t, *h, **all_V, **av, **au, **vr;

	map->m_R = R;
	map->m_vrtxs = (VERTEX**)malloc(R * sizeof(VERTEX*));

#ifdef CHECK_ERRORS
	if (map->m_vrtxs == NULL)
	{
		printf("Error 5: Not enough memory.\n");
		exit(-5);
	}
#endif

	all_V = (VERTEX**)malloc(graph->m_V * sizeof(VERTEX*));

#ifdef CHECK_ERRORS
	if (all_V == NULL)
	{
		printf("Error 6: Not enough memory.\n");
		exit(-6);
	}
#endif

	av = all_V;

	for (v = graph->m_vrtxs, t = graph->m_vrtxs + graph->m_V; v < t; ++v)
	{
		*av = v;
		++av;
	}

	k = graph->m_V;
	av = all_V;
	vr = map->m_vrtxs;

	for (i = 0; i < R; ++i)
	{
		r = random() % k;
		au = av + r;

		h = *au;
		*au = *av;
		*vr = *av = h;

		++av;
		++vr;
		--k;
	}
}


void destroy_mapping(MAPPING *map)
{
	free(map->m_vrtxs);
}


void read_umapping_start(char *filename, UGRAPH *ugraph)
{
	FILE *fr;

	int R;
	fr = fopen(filename, "r");

#ifdef CHECK_ERRORS
	if (fr == NULL)
	{
		printf("Error 12: Cannot open file %s.\n", filename);
		exit(-12);
	}
#endif

	unsigned int i, x, y;

	fscanf(fr, "p robot %u\n", &R);

	for (i = 0; i < R; ++i)
	{
		fscanf(fr, "r %u %u\n", &x, &y);

		UVERTEX *v = ugraph->m_root;

		while (v != NULL)
		{
			if (v->m_id == y)
			{
				v->m_rob = x;
				break;
			}
			v = v->m_next;
		}
	}

	fclose(fr);
}


void read_umapping_goal(char *filename, UGRAPH *ugraph)
{
	FILE *fr;

	int R;
	fr = fopen(filename, "r");

#ifdef CHECK_ERRORS
	if (fr == NULL)
	{
		printf("Error 12: Cannot open file %s.\n", filename);
		exit(-12);
	}
#endif

	unsigned int i, x, y;

	fscanf(fr, "p robot %u\n", &R);

	for (i = 0; i < R; ++i)
	{
		fscanf(fr, "r %u %u\n", &x, &y);

		UVERTEX *v = ugraph->m_root;

		while (v != NULL)
		{
			if (v->m_id == y)
			{
				v->m_gob = x;
				break;
			}
			v = v->m_next;
		}
	}

	fclose(fr);
}


unsigned int ok_A, tok_A;

void initialize(MAPPING *init, MAPPING *goal, GRAPH *graph)
{
	unsigned int i, r;

	for (i = 0; i < graph->m_V; ++i)
	{
		graph->m_vrtxs[i].m_val = -1;
	}

	for (r = 0; r < init->m_R; ++r)
	{
		init->m_vrtxs[r]->m_val = r;
	}

	tok_A = init->m_R;
	ok_A = 0;

	for (r = 0; r < init->m_R; ++r)
	{
		if (init->m_vrtxs[r] == goal->m_vrtxs[r])
		{
			++ok_A;
		}
	}
}


int search_A(unsigned int d, unsigned int md, MAPPING *map, MAPPING *goal, GRAPH *graph)
{
	++STAT_steps;

	unsigned int r;

	if (ok_A >= tok_A)
	{
		return 1;
	}
	else
	{
		if (d > md)
		{
			return 0;
		}
		else
		{
			for (r = 0; r < map->m_R; ++r)
			{
				VERTEX **n = map->m_vrtxs[r]->m_nghs;

				while (*n != NULL)
				{
					if ((*n)->m_val == -1)
					{
						VERTEX *s = map->m_vrtxs[r];
						
						map->m_vrtxs[r] = *n;
						(*n)->m_val = r;
						s->m_val = -1;
/*						
						sol_A[d].m_from = s;
						sol_A[d].m_to = *n;
*/					
						unsigned int ook = ok_A;

						if (goal->m_vrtxs[r] == *n)
						{
							++ok_A;
						}
						else
						{
							if (goal->m_vrtxs[r] == s)
							{
								--ok_A;
							}
						}

						if (search_A(d + 1, md, map, goal, graph) == 1)
						{
							return 1;
						}

						ok_A = ook;
						
						(*n)->m_val = -1;
						s->m_val = r;
						map->m_vrtxs[r] = s;
					}
					++n;
				}
			}

			return 0;
		}
	}
}


int get_quality(MAPPING *map, MAPPING *goal)
{
	unsigned int r, q = 0;

	for (r = 0; r < map->m_R; ++r)
	{
		if (map->m_vrtxs[r] == goal->m_vrtxs[r])
		{
			++q;
		}
	}

	return q;
}


int get_quality_dist(MAPPING *map, MAPPING *goal, GRAPH *graph)
{
	unsigned int r, q = 0;

	for (r = 0; r < map->m_R; ++r)
	{
		unsigned int u = map->m_vrtxs[r] - graph->m_vrtxs;
		unsigned int v = goal->m_vrtxs[r] - graph->m_vrtxs;

		q += dist_B.m_d[u][v];
	}

	return q;
}


int search_B(unsigned int d, unsigned int md, MAPPING *map, MAPPING *goal, GRAPH *graph)
{
	++STAT_steps;

	unsigned int r;

	if (ok_A >= tok_A)
	{
		return 1;
	}
	else
	{
		if (d > md)
		{
			return 0;
		}
		else
		{
			int q;

//			for (q = map->m_R; q >= 0; --q)
//			for (q = map->m_R * (graph->m_V + 1); q >= 0; --q)
			for (q = 0; q <= map->m_R * (graph->m_V + 1); ++q)
			{
				for (r = 0; r < map->m_R; ++r)
				{
					VERTEX **n = map->m_vrtxs[r]->m_nghs;

					while (*n != NULL)
					{
						if ((*n)->m_val == -1)
						{
							VERTEX *s = map->m_vrtxs[r];
						
							map->m_vrtxs[r] = *n;
							(*n)->m_val = r;
							s->m_val = -1;
						
							if (get_quality_dist(map, goal, graph) == q)
//							if (get_quality(map, goal) == q)
							{
/*
								sol_A[d].m_from = s;
								sol_A[d].m_to = *n;
*/
						
								unsigned int ook = ok_A;

								if (goal->m_vrtxs[r] == *n)
								{
									++ok_A;
								}
								else
								{
									if (goal->m_vrtxs[r] == s)
									{
										--ok_A;
									}
								}
								
//								if (ok_A == q)
								{
									if (search_B(d + 1, md, map, goal, graph) == 1)
									{
										return 1;
									}
								}

								ok_A = ook;
							}
							
							(*n)->m_val = -1;
							s->m_val = r;
							map->m_vrtxs[r] = s;
						}
						++n;
					}
				}
			}

			return 0;
		}
	}
}


int is_solved(MAPPING *map, MAPPING *goal)
{
	unsigned int r;

	for (r = 0; r < map->m_R; ++r)
	{
		if (map->m_vrtxs[r] != goal->m_vrtxs[r])
		{
			return 0;
		}
	}

	return 1;
}


void solve_A(MAPPING *init, MAPPING *goal, GRAPH *graph)
{
	unsigned int md = 1;

	initialize(init, goal, graph);

	while (1)
	{
		printf("Depth: %u ...\n", md);

		if (search_A(0, md, init, goal, graph) == 1)
		{
			printf("Solved in depth %u.\n", md);

			unsigned int i;

			for (i = 0; i <= md; ++i)
			{
				printf("%u->%u ", sol_A[i].m_from - graph->m_vrtxs, sol_A[i].m_to - graph->m_vrtxs);
			}
			printf("\n");

			break;
		}
		++md;
	}
}


void solve_B(MAPPING *init, MAPPING *goal, GRAPH *graph)
{
	unsigned int md = 1;

	initialize(init, goal, graph);
	setup_distances(graph, &dist_B);
	all_pairs_shortest(&dist_B);

	while (1)
	{
		printf("Depth: %u ...\n", md);

		if (search_B(0, md, init, goal, graph) == 1)
		{
			printf("Solved in depth %u.\n", md);

			unsigned int i;

			for (i = 0; i <= md; ++i)
			{
				printf("%u->%u ", sol_A[i].m_from - graph->m_vrtxs, sol_A[i].m_to - graph->m_vrtxs);
			}
			printf("\n");

			break;
		}
		++md;
	}
}


GRAPH g, g2, g3;
MAPPING m, m2, m3;
char goutput[10000];
char moutput[10000];

GRAPH graph;
MAPPING start, goal;
DIST dist;
UPOOL upool1, upool2;
GPOOL gpool;
CPOOL cpool;
UGRAPH ugraph;
GGRAPH tree;
USTACK ustack;

int main(void)
{
	clock_t b, e;

	b = clock();
/*
	for (i = 0; i < 4000; ++i)
	{
		create_random_graph(200, 4000, &g);
		//graph2string(&g, output);
		destroy_graph(&g);
	}
*/
/*
	create_random_graph(10, 20, &g);
	graph2string(&g, goutput);
	printf("%s\n", goutput);

	create_random_mapping(5, &g, &m);
	mapping2string(&g, &m, moutput);
	printf("%s\n", moutput);

	destroy_graph(&g);
	destroy_mapping(&m);

	string2mapping("p robot 5\nr 0 2\nr 1 3\nr 3 1\nr 2 6\nr 4 5\n", &g, &m2);

	mapping2string(&g, &m2, moutput);
	printf("%s\n", moutput);

	string2graph("p edge 5 4\ne 0 1\ne 0 2\ne 1 2\ne2 3\n", &g2);
	graph2string(&g2, goutput);
	printf("%s\n", goutput);
*/
/*
	read_graph("graph_01.col", &g3);
	graph2string(&g3, goutput);
	printf("%s\n", goutput);

	read_mapping("mapping_01.map", &g3, &m3);
	mapping2string(&g3, &m3, moutput);
	printf("%s\n", moutput);

	write_graph("graph_02.col", &g3);
	write_mapping("mapping_02.map", &g3, &m3);
*/
/*
	read_graph("graph_A_07.col", &graph);
	print_graph(&graph);
	decompose_biconnected(&graph);
	print_graph(&graph);
*/
/*
	graph2string(&graph, goutput);
	printf("%s\n", goutput);

	read_mapping("start_A_02.map", &graph, &start);
	mapping2string(&graph, &start, moutput);
	printf("%s\n", moutput);

	read_mapping("goal_A_02.map", &graph, &goal);
	mapping2string(&graph, &goal, moutput);
	printf("%s\n", moutput);
	
	init_star(&start, &goal, &graph);
	print_distances(&dist_B);
	search_star(&goal, &graph);
*/
/*
	init_upool(&upool, 100, 10000);
	init_ugraph(&ugraph, &upool);

	add_vertex_ugraph(1, &ugraph, &upool);
	add_vertex_ugraph(2, &ugraph, &upool);
	add_vertex_ugraph(3, &ugraph, &upool);
	add_vertex_ugraph(4, &ugraph, &upool);

	add_edge_ugraph(1, 2, &ugraph, &upool);
	add_edge_ugraph(2, 3, &ugraph, &upool);
	add_edge_ugraph(3, 4, &ugraph, &upool);
	add_edge_ugraph(4, 1, &ugraph, &upool);
	add_edge_ugraph(1, 3, &ugraph, &upool);

	print_ugraph(&ugraph);
*/

	read_ugraph("graph_cross_40.col", &ugraph, &upool1);
	read_umapping_start("start_cross_40.map", &ugraph);
	read_umapping_goal("goal_cross_40.map", &ugraph);

	printf("--> Ksi 1\n");
	decompose_ubiconnected(&ugraph, &ustack, &tree, &upool2, &gpool, &cpool);
	printf("--> Ksi 2\n");
	print_ugraph(&ugraph);
	printf("--> Ksi 3\n");
	print_ggraph(&tree);
	printf("--> Ksi 4\n");
	solve_ggraph(&tree);
	printf("--> Ksi 5\n");

//	write_ugraph("graph_A_07c.col", &ugraph);

	e = clock();
	printf("Time:%.3f\n", (e - b) / (double)CLOCKS_PER_SEC);
	printf("Steps:%llu\n", STAT_steps);

	return 0;
}
